#!/usr/bin/env python3


"""Define the table data structure.

Author: zjl9959@gmail.com

License: GPLv3

"""

from algviz.svg_table import SvgTable
from algviz.cursor import Cursor, _CursorManager
from algviz.utility import TraceColorStack, AlgvizParamError, clamp
from algviz.utility import kMinAnimDelay, kMaxAnimDelay, kMinCellWidth
from algviz.utility import kMaxCellWidth, kMaxBarHight, kMinCellHeight, kMaxCellHeight


class Vector():
    """
    A vector is an expandable list, you can add or remove cell for vector.
    When the cell position changed, the vector display SVG will show the related animation.
    
    [WARNING] Don't create this class directly, use algviz.Visualizer.createTable instead.
    """

    def __init__(self, data, delay, cell_size, histogram, show_index):
        """
        Args:
            data (list(printable)): The initialize data for vector.
            delay (float): Animation delay time between two animation frames.
            cell_size tuple(float, float): Vector cell size (width, height).
            histogram (bool): Display the data in the form of a histogram or not.
            show_index (bool): Whether to display the vector index label.
        """
        if type(cell_size) != tuple or len(cell_size) < 2:
            raise AlgvizParamError('Vector cell_size parameter should be <tuple(float, float)> type.')
        self._data = list()             # Store the data in vector list.
        if data is not None:
            for i in range(len(data)):
                self._data.append(data[i])
        self._delay = clamp(delay, kMinAnimDelay, kMaxAnimDelay)                 # Animation delay time.
        self._cell_width = clamp(cell_size[0], kMinCellWidth, kMaxCellWidth)
        if histogram:
            self._cell_height = clamp(cell_size[1], kMinCellHeight, kMaxBarHight)
        else:
            self._cell_height = clamp(cell_size[1], kMinCellHeight, kMaxCellHeight)
        self._show_histogram = histogram          
        self._show_index = show_index   # Whether to display the vector index label.
        self._cell_margin = 3           # Margin between two adjacent cells.
        self._cell_tcs = dict()         # Record the trajectory access information (node_index: ColorStack) of all cells.
        self._frame_trace_old = list()  # Cache the cell related information that needs to be cleared in the previous frame.
        self._frame_trace = list()      # Record the relevant information of the cell to be refreshed in the next frame.
        self._rect_move = dict()        # Record the index of moving cells and it's relative moving distance in the next frame.
        self._rect_disappear = list()   # Record the index of disappearing cells in the next frame.
        self._rect_appear = list()      # Record the index of appearing cells in the next frame.
        self._index2rect = dict()       # The mapping relationship from vector index to the cell object.
        self._index2text = dict()       # The mapping relationship from vector index to the text object.
        self._label_font_size = int(min(12, self._cell_width*0.5))   # The font size of the vector's subscript index.
        self._next_iter = 0             # Mark the positon of current iteration.
        self._svg = SvgTable(self._cell_margin, self._cell_margin)
        # Initial cursor manager.
        self._cursor_manager = _CursorManager((self._cell_width, self._cell_width), self._svg, 'D',
            (self._cell_margin, self._cell_margin), self._cell_margin)
        # Create rect elements for initial data.
        for i in range(len(self._data)):
            rect_pos_x = self._cell_width*i + self._cell_margin*(i+1)
            rect_pos_y = self._cell_margin + self._cursor_manager.get_cursors_occupy()
            rect = (rect_pos_x, rect_pos_y, self._cell_width, self._cell_height)
            rid = self._svg.add_rect_element(rect, text=self._data[i])
            self._cell_tcs[rid] = TraceColorStack()
            self._index2rect[i] = rid
        # Update SVG and rects size.
        self._update_svg_size_(len(self._data))
        if self._show_histogram:
            self._update_bar_height_()
        if self._show_index:
            self._create_new_subscripts_(0, len(self._data))


    def insert(self, index, val):
        """Insert a new value into vector. If index < 0 or index >= length of Vector, then set index = index % vector length.
        
        Args:
            index (int/Cursor): The subscript index to insert value. (Insert before index)
            val (printable): The value to insert into vector.

        Raises:
            RuntimeError: Index:xxx type is not int or Cursor.
            RuntimeError:  Vector index=xxx out of range!
        """
        index = self._check_index_type_and_range_(index)
        # Add a new rectangle node and animation to SVG.
        rect_pos_x = self._cell_width*index+self._cell_margin*(index+1)
        rect_pos_y = self._cell_margin + self._cursor_manager.get_cursors_occupy()
        rect = (rect_pos_x, rect_pos_y, self._cell_width, self._cell_height)
        rid = self._svg.add_rect_element(rect, text=val)
        # Record the cells need to be move after the insert postion.
        for i in range(len(self._data), index, -1):
            rrid = self._index2rect[i-1]
            self._index2rect[i] = rrid
            if rrid in self._rect_move:
                self._rect_move[rrid] += 1
            else:
                self._rect_move[rrid] = 1
        self._index2rect[index] = rid
        self._cell_tcs[rid] = TraceColorStack()
        self._rect_appear.append(rid)
        self._data.insert(index, val)
    

    def append(self, val):
        """Append a new value into vector's tail.
        
        Args:
            val (printable): The value to appended into vector's tail.
        """
        index = len(self._data)
        rect_pos_x = self._cell_width*index+self._cell_margin*(index+1)
        rect_pos_y = self._cell_margin + self._cursor_manager.get_cursors_occupy()
        rect = (rect_pos_x, rect_pos_y, self._cell_width, self._cell_height)
        rid = self._svg.add_rect_element(rect, text=val)
        self._index2rect[index] = rid
        self._cell_tcs[rid] = TraceColorStack()
        self._rect_appear.append(rid)
        self._data.append(val)
    

    def pop(self, index=None):
        """Pop a value from vector. Pop vector's tail value as default. 
        
        Args:
            index (int/Cursor): The index position of value to pop out.

        Raises:
            RuntimeError: Index:xxx type is not int or Cursor.
            RuntimeError:  Vector index=xxx out of range!
        """
        if index == None:
            index = len(self._data) - 1
        else:
            index = self._check_index_type_and_range_(index)
        rid = self._index2rect[index]
        for i in range(index, len(self._data)-1):
            rrid = self._index2rect[i+1]
            self._index2rect[i] = rrid
            if rrid in self._rect_move:
                self._rect_move[rrid] -= 1
            else:
                self._rect_move[rrid] = -1
        self._index2rect.pop(len(self._data)-1)
        self._rect_disappear.append(rid)
        return self._data.pop(index)


    def clear(self):
        """Clear all the values in vector.
        """
        for i in range(len(self._data)):
            rid = self._index2rect[i]
            self._rect_disappear.append(rid)
        self._index2rect.clear()
        self._rect_move.clear()
        self._rect_appear.clear()
        self._data.clear()
    

    def swap(self, index1, index2):
        """Swap the two cells positon in Vector.
        Args:
            index1, index2 (int/Cursor): The two index positions to be swapped.
        
        Raises:
            RuntimeError: Index:xxx type is not int or Cursor.
            RuntimeError:  Vector index=xxx out of range!
        """
        index1 = self._check_index_type_and_range_(index1)
        index2 = self._check_index_type_and_range_(index2)
        rid1 = self._index2rect[index1]
        rid2 = self._index2rect[index2]
        self._index2rect[index1] = rid2
        self._index2rect[index2] = rid1
        if rid1 in self._rect_move.keys():
            self._rect_move[rid1] += index2 - index1
        else:
            self._rect_move[rid1] = index2 - index1
        if rid2 in self._rect_move.keys():
            self._rect_move[rid2] += index1 - index2
        else:
            self._rect_move[rid2] = index1 - index2
        temp_data = self._data[index2]
        self._data[index2] = self._data[index1]
        self._data[index1] = temp_data


    def mark(self, color, st, ed=None, hold=False):
        """Emphasize one cell in the Vector by mark it's background color.
            
        Args:
            color ((R,G,B)): The background color for the marked cell. R, G, B stand for color channel for red, green, blue.
                R,G,B should be int value and 0 <= R,G,B <= 255. eg:(0, 255, 0)
            st, ed (int/Cursor): The mark range's index in Vector.
            hold (bool): Whether to keep the mark color in future animation frames.
        """
        st = self._check_index_type_and_range_(st)
        if ed is None:
            ed = st + 1
        if type(ed) is Cursor:
            ed = ed.index()
        for i in range(st, ed):
            if i < 0 or i >= len(self._data):
                i %= len(self._data)
            rid = self._index2rect[i]
            self._cell_tcs[rid].add(color)
            self._frame_trace.append((rid, color, hold))
    

    def removeMark(self, color):
        """Remove the mark color for cell(s).
        
        Args:
            color ((R,G,B)): R, G, B stand for color channel for red, green, blue.
                R,G,B should be int value and 0 <= R,G,B <= 255. eg:(0, 255, 0)
        """
        for rid in self._index2rect.values():
            if self._cell_tcs[rid].remove(color):
                self._svg.update_rect_element(rid, fill=self._cell_tcs[rid].color())
    

    def _add_cursor_(self, cursor):
        """Create a new cursor to track the element's index.
        
        Args:
            cursor (Cursor): The cursor object to track.
        
        Returns:
            bool: Wheather add cursor successfully.
        """
        if type(cursor) != Cursor:
            return False
        if not self._cursor_manager.contains(cursor):
            self._cursor_manager.add_cursor(cursor)
            self._update_svg_size_(len(self._data))
            self._update_rects_position_()
            self._update_subscripts_position_()
            return True
        return False


    def _remove_cursor_(self, cursor):
        """Remove one cursor from Vector and it's SVG representation.

        Args:
            cursor (Cursor): The cursor object to be removed.

        Returns:
            bool: Wheather remove cursor successfully.
        """
        if type(cursor) != Cursor:
            return False
        if self._cursor_manager.contains(cursor):
            self._cursor_manager.remove_cursor(cursor)
            self._update_svg_size_(len(self._data))
            self._update_rects_position_()
            self._update_subscripts_position_()
            return True
        return False


    def __getitem__(self, index):
        """
        Args:
            index (int/Cursor): The index position of the cell to be accessed.

        Raises:
            RuntimeError: Index:xxx type is not int or Cursor.
            RuntimeError:  Vector index=xxx out of range!
        """
        if type(index) is Cursor:
            self._add_cursor_(index)
        index = self._check_index_type_and_range_(index)
        return self._data[index]
    

    def __setitem__(self, index, val):
        """
        Args:
            index (int/Cursor): The index position of the cell to be updated.
            val (printable): New value for the cell.
        
        Raises:
            RuntimeError: Index:xxx type is not int or Cursor.
            RuntimeError:  Vector index=xxx out of range!
        """
        if type(index) is Cursor:
            self._add_cursor_(index)
        index = self._check_index_type_and_range_(index)
        rid = self._index2rect[index]
        label = val
        if val is None:
            label = ''
        self._svg.update_rect_element(rid, text=label)
        self._data[index] = val
    

    def __len__(self):
        """
        Returns:
            int: Length of Vector.
        """
        return len(self._data)
    
    def __iter__(self):
        self._next_iter = 0
        return self
    
    def __next__(self):
        if self._next_iter >= len(self._data):
            raise StopIteration
        else:
            res = self[self._next_iter]
            self._next_iter += 1
            return res
    

    def _repr_svg_(self):
        """
        Returns:
            str: The SVG representation of current Vector.
        """
        # Update the color of the cell tracker.
        all_data_num = len(self._data) + len(self._rect_disappear)
        self._update_svg_size_(all_data_num)
        for (rid, color) in self._frame_trace_old:
            if rid not in self._cell_tcs.keys():
                continue
            if (rid, color, False) not in self._frame_trace and (rid, color, True) not in self._frame_trace:
                self._cell_tcs[rid].remove(color)
            self._svg.update_rect_element(rid, fill=self._cell_tcs[rid].color())
        self._frame_trace_old.clear()
        for (rid, color, hold) in self._frame_trace:
            self._svg.update_rect_element(rid, fill=self._cell_tcs[rid].color())
            if not hold:
                self._frame_trace_old.append((rid, color))
        self._frame_trace.clear()
        # Add animations of the disappearance of cells.
        disappear_animate_end_time = self._delay*0.2
        for rid in self._rect_disappear:
            self._svg.add_animate_appear(rid, (0, disappear_animate_end_time), appear=False)
        has_move_animate = False
        move_animate_start_time = disappear_animate_end_time if len(self._rect_disappear) else 0
        move_animate_end_time = move_animate_start_time + (self._delay-move_animate_start_time)*0.6
        # Add animations of cells movement.
        if self._show_histogram > 0:
            self._update_bar_height_()
        for rid in self._rect_move.keys():
            if self._rect_move[rid] == 0:
                continue
            self._svg.add_animate_move(rid, (self._rect_move[rid]*(self._cell_width+self._cell_margin), 0),
                                        (move_animate_start_time, move_animate_end_time), bessel=False)
            has_move_animate = True
        # Add animations of the appearance of cells.
        appear_animate_start_time = move_animate_end_time if has_move_animate else move_animate_start_time
        for rid in self._rect_appear:
            self._svg.add_animate_appear(rid, (appear_animate_start_time, self._delay))
        # Update subscript index of vector cells.
        if self._show_index:
            if len(self._index2text) > len(self._data):
                for i in range(len(self._data), len(self._index2text)):
                    self._svg.delete_element(self._index2text[i])
                    self._index2text.pop(i)
            elif len(self._index2text) < len(self._data):
                self._create_new_subscripts_(len(self._index2text), len(self._data))
        self._rect_move.clear()
        self._cursor_manager.refresh_cursors_animation(all_data_num, (0, self._delay))
        res = self._svg._repr_svg_()
        # Clear the animation effect, update the SVG content, and prepare for the next frame.
        self._svg.clear_animates()
        if self._show_histogram > 0:
            self._update_bar_height_()
        else:
            for i in range(len(self._data)):
                rect_pos_x = self._cell_width*i + self._cell_margin*(i+1)
                rect_pos_y = self._cell_margin + self._cursor_manager.get_cursors_occupy()
                rect = (rect_pos_x, rect_pos_y, self._cell_width, self._cell_height)
                rid = self._index2rect[i]
                self._svg.update_rect_element(rid, rect=rect)
        for rid in self._rect_disappear:
            self._svg.delete_element(rid)
            self._cell_tcs.pop(rid)
        self._rect_disappear.clear()
        for rid in self._rect_appear:
            self._svg.update_rect_element(rid, opacity=True)
        self._rect_appear.clear()
        self._cursor_manager.update_cursors_position()
        return res
    

    def _update_bar_height_(self):
        """Update the height of each column in the histogram.
        """
        # Adjust the ratio and baseline position according to the value numbers range.
        mmax_data, max_data = 0, 0
        for num in self._data:
            if num is None:
                continue
            num = float(num)
            if num < 0:
                mmax_data = min(mmax_data, num)
            else:
                max_data = max(max_data, num)
        if (max_data - mmax_data) < 0.0001:
            ratio = 0
        else:
            useful_height = self._cell_height - 2*self._cell_margin
            if self._show_index:
                useful_height -= self._label_font_size
            ratio = useful_height/(max_data-mmax_data)
        baseline = max_data*ratio + self._cell_margin + self._cursor_manager.get_cursors_occupy()
        # Update the position coordinates of cells.
        for i in range(len(self._data)):
            if self._data[i] is None:
                num = 0
            else:
                num = float(self._data[i])
            rid = self._index2rect[i]
            x = self._cell_width*i + self._cell_margin*(i+1)
            if rid in self._rect_move.keys():
                x -= self._rect_move[rid] * (self._cell_width+self._cell_margin)
            height = ratio*num
            if num < 0:
                y = baseline
            else:
                y = baseline - height
            if num - int(num) > 0.001:
                num = '{:.2f}'.format(num)
            else:
                num = '{:.0f}'.format(num)
            if self._data[i] is None:
                num = None
            self._svg.update_rect_element(rid, rect=(x, y, self._cell_width, abs(height)), text=num)


    def _update_svg_size_(self, data_num):
        if self._show_histogram:
            self._svg_height = self._cell_height
        else:
            self._svg_height = self._cell_height + 2*self._cell_margin
            if self._show_index:
                self._svg_height += self._label_font_size
        self._svg_height += self._cursor_manager.get_cursors_occupy()
        self._svg_width = data_num*self._cell_width+(data_num+1)*self._cell_margin
        self._svg.update_svg_size(self._svg_width, self._svg_height)

    def _update_rects_position_(self):
        for i, gid in self._index2rect.items():
            rect_pos_x = self._cell_width*i+self._cell_margin*(i+1)
            rect_pos_y = self._cell_margin + self._cursor_manager.get_cursors_occupy()
            rect = (rect_pos_x, rect_pos_y, self._cell_width, self._cell_height)
            if self._svg:
                self._svg.update_rect_element(gid, rect)

    def _create_new_subscripts_(self, st, ed):
        for i in range(st, ed):
            pos_x = self._cell_width*(i+0.5)+self._cell_margin*(i+1)-self._label_font_size*len(str(i))*0.25
            pos_y = self._svg_height - self._cell_margin
            tid = self._svg.add_text_element((pos_x, pos_y), i, font_size=self._label_font_size)
            self._index2text[i] = tid

    def _update_subscripts_position_(self):
        for i, gid in self._index2text.items():
            gid = self._index2text[i]
            pos_x = self._cell_width*(i+0.5)+self._cell_margin*(i+1)-self._label_font_size*len(str(i))*0.25
            pos_y = self._svg_height - self._cell_margin
            self._svg.update_text_element(gid, (pos_x, pos_y))

    def _check_index_type_and_range_(self, index):
        res = None
        if type(index) is int:
            res = index
        elif type(index) is Cursor:
            res = index.index()
        else:
            raise RuntimeError('Index:{} type is not int or Cursor.'.format(index))
        if index < 0 or index >= len(self._data):
            raise RuntimeError('Vector index={} out of range!'.format(index))
        return res



#!/usr/bin/env python3


"""Define the cursor related classes.

A cursor is an arrow shown in Vector or Table's SVG display.
The cursor represents the index of the element in Vector or Table.

Author: zjl9959@gmail.com

License: GPLv3

"""

from algviz.utility import AlgvizTypeError


# The alternative cursor colors list.
kcursor_colors = (
    (40, 116, 166),
    (147, 81, 22),
    (25, 111, 61),
    (203, 67, 53),
    (28, 40, 51),
    (99, 57, 116)
)


class Cursor:
    """A cursor is a display object to show the current index of Vector or Table(row/column).

    A cursor object should be created from Visualizer.create_cursor
    or Visualizer.cursorRange interfaces. After that,
    you can access the element in any Vector/Table object with
    the created cursor just like an integer index.
    The index in the cursor should be an integer number.
    
    Assignment operation: <<

    Mathmatics operations:
        No side effect, returns a int number: +, -, *, //, %

        Changes the cursor's index and returns itself: +=, -=, *=, //=, %=

    Compare operations: >, <, >=, <=, ==, !=
    
    The right-hand value of those operations above can be a cursor or an integer number.


    For example:
        i = viz.createCursor(3, 'i')   # The index in cursor i is 3.

        j = viz.createCursor(5, 'j')   # The index in cursor j is 5.

        i += j                          # The index in cursor i change from 3 into 8(3+5).

        j << i - 1                      # The index in cursor j change from 5 into 7(8-1).
        
        i > j                           # True: 8 > 7 is true.
    """

    def __init__(self, name, index, id):
        self._managers = list()         # This cursor's related managers.
        self._name = name               # The cursor's display name.
        self._index = index             # The current index of cursor.
        color = kcursor_colors[id % len(kcursor_colors)]
        self._color = color             # The display color of cursor.

    def _add_manager_(self, mgr):
        if mgr not in self._managers:
            self._managers.append(mgr)

    def _on_cursor_updated_(self, new_index):
        for mgr in self._managers:
            mgr.on_cursor_updated(self, new_index)

    def index(self):
        """
        Returns:
            int: The cursor's current index value.
        """
        return self._index

    def name(self):
        """
        Returns:
            str: Cursor display name.
        """
        return self._name

    # Index assignment operator.
    def __lshift__(self, other):
        self._index = _get_rhs_index(other)
        self._on_cursor_updated_(self._index)
        return self

    # Mathmatics operators.
    def __add__(self, other):
        rhs = _get_rhs_index(other)
        return self._index + rhs

    def __sub__(self, other):
        rhs = _get_rhs_index(other)
        return self._index - rhs

    def __mul__(self, other):
        rhs = _get_rhs_index(other)
        return self._index * rhs

    def __floordiv__(self, other):
        rhs = _get_rhs_index(other)
        return self._index // rhs

    def __mod__(self, other):
        rhs = _get_rhs_index(other)
        return self._index % rhs

    def __imul__(self, other):
        self._index *= _get_rhs_index(other)
        self._on_cursor_updated_(self._index)
        return self

    def __ifloordiv__(self, other):
        self._index //= _get_rhs_index(other)
        self._on_cursor_updated_(self._index)
        return self

    def __iadd__(self, other):
        self._index += _get_rhs_index(other)
        self._on_cursor_updated_(self._index)
        return self

    def __isub__(self, other):
        self._index -= _get_rhs_index(other)
        self._on_cursor_updated_(self._index)
        return self

    def __imod__(self, other):
        self._index %= _get_rhs_index(other)
        self._on_cursor_updated_(self._index)
        return self

    # Compare operators.
    def __lt__(self, other):
        return self._index < _get_rhs_index(other)

    def __le__(self, other):
        return self._index <= _get_rhs_index(other)

    def __eq__(self, other):
        return self._index == _get_rhs_index(other)

    def __ne__(self, other):
        return self._index != _get_rhs_index(other)

    def __gt__(self, other):
        return self._index > _get_rhs_index(other)

    def __ge__(self, other):
        return self._index >= _get_rhs_index(other)


def _get_rhs_index(rhs):
    if rhs == None:
        return None
    elif type(rhs) is int:
        return rhs
    elif type(rhs) is Cursor:
        return rhs._index
    else:
        raise AlgvizTypeError(rhs)


class _CursorRange:
    def __init__(self, id, name, start, end, step):
        self._end = end
        self._step = step
        self._next_index = start
        self._cursor = Cursor(name, start, id)

    def __iter__(self):
        return self

    def __next__(self):
        if self._next_index == self._end:
            raise StopIteration
        else:
            self._cursor << self._next_index
            self._next_index += self._step
            return self._cursor


class _CursorManager:
    def __init__(self, cell_size, svg, dir, svg_margin=(0, 0), cell_margin=0):
        """
        Args:
            cell_size (float, float): The rectangle cell size in SVG.
            svg (SvgTable): The SvgTable object to add cursors on.
            dir (str): The direction of cursors. (Support 'D':down, 'R':right)
            svg_margin (float, float): The margin between SVG side and cursor's start.
            cell_margin (float): The margin between rectangle cells.
        """
        self._next_cursor_id = 0
        self._cell_width = cell_size[0]
        self._cell_height = cell_size[1]
        self._svg_margin = svg_margin           # The margin between cell and SVG.
        self._cell_margin = cell_margin         # The margin between cells.
        self._svg = svg                         # The SVGTable object in Vector or Table.
        self._dir = dir                         # The direction of the cursor's arrow (U:up, D:down, L:left, R:right).
        # The total height of a cursor (arrow and label).
        if self._dir == 'U' or self._dir == 'D':
            self._cursor_height = min(20, 0.4 * self._cell_height)
        else:
            self._cursor_height = min(20, 0.4 * self._cell_width)
        self._cursor_margin = 3                 # The margin between cursor manager and SVG.
        self._cursor2id = dict()                # Map the cursor object id into the unique cursor id in this cursor manager.
        # Cursor_offset = cursor_id*self._cursor_offset.
        self._cursor_offset = 2
        self._cursors_info = dict()             # key:cursor_id; value:(cursor_gid, cursor_color, cursor_name) in SVG.
        self._cursors_id_list = list()          # Keep the sequence of cursors id by create time.
        self._old_cursors_index = dict()        # key:cursor_id; value:cursor_index.
        self._new_cursors_index = dict()        # key:cursor_id; value:cursor_index.
        self._cursor_moves = dict()             # Cache the cursor move since last frame. key:cursor_id; value:(cursor_move_delt_x, cursor_move_delt_y).

    def contains(self, cursor):
        """Check if the cursor was tracked by this manager.
        """
        return id(cursor) in self._cursor2id

    def add_cursor(self, cursor):
        """Create a new cursor object and track it.
        """
        if self.contains(cursor):
            return
        name = cursor.name()
        index = cursor.index()
        # Update old cursor's position.
        for cid in self._cursors_id_list:
            gid = self._cursors_info[cid][0]
            if self._dir == 'R':
                self._svg.update_cursor_element(gid, (self._cursor_height, 0))
            else:
                self._svg.update_cursor_element(gid, (0, self._cursor_height))
        # Create a new cursor and setup it's position and color.
        cursor_id = self._next_cursor_id
        self._next_cursor_id += 1
        cursor_seq = len(self._cursors_id_list)
        cursor_pos = self._calculate_cursor_position_(cursor_seq, index)
        cursor_color = cursor._color
        cursor_node = self._svg.add_cursor_element(cursor_pos, cursor_color, name, self._dir)
        # Record the cursor's position and offset information.
        self._cursors_id_list.append(cursor_id)
        self._cursors_info[cursor_id] = (cursor_node, cursor_color, name)
        self._old_cursors_index[cursor_id] = index
        # Update the cache and cursor.
        self._cursor2id[id(cursor)] = cursor_id
        cursor._add_manager_(self)


    def remove_cursor(self, cursor):
        """Untrack the specific cursor from this cursor manager.
        """
        if id(cursor) not in self._cursor2id:
            return
        cursor_id = self._cursor2id[id(cursor)]
        if cursor_id not in self._cursors_id_list:
            return
        cursor_seq = self._cursors_id_list.index(cursor_id)
        # Update the cursors who's sequence ahead of the cursor to be removed.
        for i in range(cursor_seq):
            cid = self._cursors_id_list[i]
            if cid not in self._cursors_info:
                print('[WARNING] _CursorManager.remove_cursor cursor {} not found in cursors_info.'.format(cid))
                continue
            gid = self._cursors_info[cid][0]
            if self._dir == 'R':
                self._svg.update_cursor_element(gid, (-self._cursor_height, 0))
            else:
                self._svg.update_cursor_element(gid, (0, -self._cursor_height))
        # Update the cursors who's sequence behind the cursor to be removed.
        for i in range(cursor_seq+1, len(self._cursors_id_list)):
            cid = self._cursors_id_list[i]
            if cid not in self._cursors_info:
                print('[WARNING] _CursorManager.remove_cursor cursor {} not found in cursors_info.'.format(cid))
                continue
            cursor_info = self._cursors_info[cid]
            # The SVGTable class don't provide a update cursor's information interface,
            # so we need to delete the cursor node in SVG and create a new one with updated information.
            self._svg.delete_element(cursor_info[0])
            cursor_index = 0
            if cid in self._old_cursors_index:
                cursor_index = self._old_cursors_index[cid]
            cursor_pos = self._calculate_cursor_position_(i-1, cursor_index)
            cursor_pos_correct = (len(self._cursors_id_list)-i-1)*self._cursor_height
            if self._dir == 'R':
                cursor_pos[0] += cursor_pos_correct
            else:
                cursor_pos[1] += cursor_pos_correct
            cursor_node = self._svg.add_cursor_element(cursor_pos, cursor_info[1], cursor_info[2], self._dir)
            self._cursors_info[cid] = (cursor_node, cursor_info[1], cursor_info[2])
        # Remove the cursor in SVG.
        if cursor_id in self._cursors_info:
            gid = self._cursors_info[cursor_id][0]
            self._svg.delete_element(gid)
        if cursor_id in self._old_cursors_index:
            self._old_cursors_index.pop(cursor_id)
        if cursor_id in self._new_cursors_index:
            self._new_cursors_index.pop(cursor_id)
        self._cursors_id_list.remove(cursor_id)

    def get_cursors_occupy(self):
        """
        Returns:
            float: The sum of all the cursors height.
        """
        return len(self._cursors_id_list) * self._cursor_height + self._cursor_margin

    def on_cursor_updated(self, cursor, new_index):
        """Called when cursor's index changed.
        """
        if not self.contains(cursor):
            return
        cursor_id = self._cursor2id[id(cursor)]
        self._new_cursors_index[cursor_id] = new_index

    def refresh_cursors_animation(self, max_index, time):
        """Refresh all the animations in cursors node.
        
        Args:
            max_index (int): Used to check the max value of the cursor.
            time (tuple(float, float)): (begin, end) The begin and end time of cursor move animation.
        """
        self._cursor_moves.clear()
        for cursor_id in self._new_cursors_index.keys():
            move_delt_x, move_delt_y = 0, 0
            old_index = self._old_cursors_index[cursor_id]
            if old_index < 0 or old_index >= max_index:
                # The cursor should move out of range if in strict mode.
                old_index = -1 if old_index < 0 else max_index
            new_index = self._new_cursors_index[cursor_id]
            if self._dir == 'R':
                move_delt_y = (new_index - old_index) * (self._cell_height + self._cell_margin)
            else:
                move_delt_x = (new_index - old_index) * (self._cell_width + self._cell_margin)
            cursor_gid = self._cursors_info[cursor_id][0]
            self._svg.add_animate_move(cursor_gid, (move_delt_x, move_delt_y), time, False)
            self._cursor_moves[cursor_id] = (move_delt_x, move_delt_y)
        # Update the cached data.
        for cursor_id in self._new_cursors_index:
            self._old_cursors_index[cursor_id] = self._new_cursors_index[cursor_id]

    def update_cursors_position(self):
        """Update all the cursors position to the latest state.
        """
        for cursor_id in self._new_cursors_index.keys():
            cursor_move = self._cursor_moves[cursor_id]
            cursor_gid = self._cursors_info[cursor_id][0]
            self._svg.update_cursor_element(cursor_gid, cursor_move)

    def on_svg_margin_changed(self, svg_margin):
        """Called when svg margin changed.
        Args:
            svg_margin (float, float): The margin between SVG side and cursor's start.
        """
        delt_svg_margin = (svg_margin[0] - self._svg_margin[0], svg_margin[1] - self._svg_margin[1])
        self._svg_margin = svg_margin
        for cursor_id in self._cursors_id_list:
            cursor_gid = self._cursors_info[cursor_id][0]
            self._svg.update_cursor_element(cursor_gid, delt_svg_margin)

    def _calculate_cursor_position_(self, cursor_seq, index):
        offset_sign = 1 if cursor_seq % 2 else -1
        cursor_offset = ((cursor_seq+1)//2)*self._cursor_offset*offset_sign
        cursor_pos_x, cursor_pos_y = 0, 0
        if self._dir == 'R':
            cursor_pos_x = self._cursor_height * (cursor_seq + 1) + self._cursor_margin
            cursor_pos_y = self._svg_margin[1] + index*(self._cell_height + self._cell_margin) + self._cell_height*0.5
            cursor_width = self._cell_height
        else:
            cursor_pos_x = self._svg_margin[0] + index*(self._cell_width + self._cell_margin) + self._cell_width*0.5
            cursor_pos_y = self._cursor_height * (cursor_seq + 1) + self._cursor_margin
            cursor_width = self._cell_width
        cursor_height = self._cursor_height * (cursor_seq + 1)
        return [cursor_pos_x, cursor_pos_y, cursor_offset, cursor_width, cursor_height]



#!/usr/bin/env python3

"""Define the table data structure related classes.

Author: zjl9959@gmail.com

License: GPLv3

"""

from algviz.svg_table import SvgTable
from algviz.cursor import Cursor, _CursorManager
from algviz.utility import AlgvizParamError, TraceColorStack, clamp
from algviz.utility import kMinCellWidth, kMaxCellWidth, kMinCellHeight, kMaxCellHeight


class TableRowIter():
    """An iterator that iterate on one specifies row in Table object.
    """

    def __init__(self, r, tab):
        """
        Args:
            r (int): Set which row in the table to iterate over.
            tab (Table): Bounded Table object for this Iterator.
        """
        self._r = r
        self._c = 0
        self._tab = tab
        
    def __iter__(self):
        self._c = 0
        return self
    
    def __next__(self):
        if self._c >= self._tab._col:
            raise StopIteration
        else:
            res = self._tab.getItem(self._r, self._c)
            self._c += 1
            return res


class TableRowOperator():
    """A manipulator to get/set the elements of the specified row in the table.
    """

    def __init__(self, r, tab):
        """
        Args:
            r (int/Cursor): Set which row in the table to iterate over.
            tab (Table): Bounded Table object for this Iterator.
        """
        self._r = r
        self._tab = tab
        
    def __getitem__(self, c):
        return self._tab.getItem(self._r, c)
        
    def __setitem__(self, c, val):
        self._tab.setItem(self._r, c, val)


class Table():
    """A static table with label text in table cells.
    
    [WARNING] Don't create this class directly, use algviz.Visualizer.createTable instead.
    """

    def __init__(self, row, col, data, cell_size, show_index):
        """
        Args:
            row (int): The number of rows for this table.
            col (int): The number of columns for this table.
            data (list(list(printable))): The initial data for table cells.
            cell_size tuple(float, float): Table cell size (width, height).
            show_index (bool): Whether to display table row and column labels.
        
        Raises:
            AlgvizParamError: Table row or col number should > 0.
        """
        if row <=0 or col <=0:
            raise AlgvizParamError('Table row or col number should > 0.')
        if type(cell_size) != tuple or len(cell_size) < 2:
            raise AlgvizParamError('Table cell_size parameter should be <tuple(float, float)> type.')
        self._row = row
        self._col = col
        self._cell_tcs = dict()             # Record the trajectory access information (node_index: ColorStack) of all cells.
        self._frame_trace_old = list()      # Cache the cell related information that needs to be cleared in the previous frame.
        self._frame_trace = list()          # Record the relevant information of the cell to be refreshed in the next frame.
        self._delay = 0                     # Animation frame delay time, used to adapt Visualizer class.
        self._next_row = 0                  # Used to mark which row the current iterated is on for the table.
        self._cell_width = clamp(cell_size[0], kMinCellWidth, kMaxCellWidth)    # The rectangle cells width in table.
        self._cell_height = clamp(cell_size[1], kMinCellHeight, kMaxCellHeight) # The rectangle cells width in table.
        self._cell_margin = 3               # The cell margin between cell and SVG side.
        self._show_index = show_index       # Wheather to show subscript of rows and columns index in table.
        self._label_font_size = 0           # The font size of the subscript labels in table.
        self._index2rect = dict()           # Map the row and column index into rectangle's gid in SVG.
        self._data = [[None for _ in range(self._col)] for _ in range(self._row)]
        self._svg = SvgTable(self._cell_margin, self._cell_margin)
        # Copy data into table.
        if data is not None:
            for r in range(self._row):
                for c in range(self._col):
                    try:
                        self._data[r][c] = data[r][c]
                    except:
                        self._data[r][c] = None
        # Add row and column cursor managers for table.
        self._row_cursor_mgr = _CursorManager((self._cell_width, self._cell_height),
            self._svg, 'R', (self._cell_margin, self._cell_margin), 0)
        self._col_cursor_mgr = _CursorManager((self._cell_width, self._cell_height),
            self._svg, 'D', (self._cell_margin, self._cell_margin), 0)
        # Initial rectangle elemenets in SVG.
        for r in range(self._row):
            for c in range(self._col):
                self._new_rect_in_svg_(r, c)
        # Initial the subscript labels in table.
        if self._show_index:
            self._label_font_size = int(min(12, self._cell_width/len(str(max(self._row,self._col)-1))))
            self._row_index2text = dict()       # Map the row index into the row subscript's text gid in SVG.
            self._col_index2text = dict()       # Map the column index into the row subscript's text gid in SVG.
            for r in range(self._row):
                self._new_row_index_text_in_svg_(r)
            for c in range(self._col):
                self._new_col_index_text_in_svg_(c)
        self._update_svg_size_()
    

    def mark(self, color, r, c, hold=False):
        """Emphasize one cell in the table by mark it's background color.
        
        Args:
            color ((R,G,B)): The background color for the marked cell. R, G, B stand for color channel for red, green, blue.
                R,G,B should be int value and 0 <= R,G,B <= 255. eg:(0, 255, 0)
            r, c (int/Cursor): Index the cell's raw, column in the table to be marked.
            hold (bool): Whether to keep the mark color in future animation frames.
        
        Raises:
            RuntimeError: Index:xx type is not int or Cursor.
            RuntimeError: Table index=xxx out of range.
        """
        r = self._check_index_type_and_range_(r, self._row)
        c = self._check_index_type_and_range_(c, self._col)
        gid = self._index2rect[(r, c)]
        self._cell_tcs[gid].add(color)
        self._frame_trace.append((gid, color, hold))
    

    def removeMark(self, color):
        """Remove the mark color for cell(s).

        Args:
            color ((R,G,B)): R, G, B stand for color channel for red, green, blue.
                R,G,B should be int value and 0 <= R,G,B <= 255. eg:(0, 255, 0)
        """
        for gid in self._index2rect.values():
            if self._cell_tcs[gid].remove(color):
                self._svg.update_rect_element(gid, fill=self._cell_tcs[gid].color())
    

    def getItem(self, r, c):
        """Get the cell value in the table.
        
        Args:
            r, c (int/Cursor): Index the cell's raw, column in the table.
        
        Returns:
            printable: The value in the specific cell.

        Raises:
            RuntimeError: Index:xx type is not int or Cursor.
            RuntimeError: Table index=xxx out of range.
        """
        if type(r) is Cursor:
            self._add_cursor_(r, True)
        if type(c) is Cursor:
            self._add_cursor_(c, False)
        r = self._check_index_type_and_range_(r, self._row)
        c = self._check_index_type_and_range_(c, self._col)
        return self._data[r][c]
    

    def setItem(self, r, c, val):
        """ Get the cell value in the table.
        
        Args:
            r, c (int/Cursor): Index the cell's raw, column in the table to be modified.
            val (printable): New value for the cell.

        Raises:
            RuntimeError: Index:xx type is not int or Cursor.
            RuntimeError: Table index=xxx out of range.
        """
        if type(r) is Cursor:
            self._add_cursor_(r, True)
        if type(c) is Cursor:
            self._add_cursor_(c, False)
        r = self._check_index_type_and_range_(r, self._row)
        c = self._check_index_type_and_range_(c, self._col)
        gid = self._index2rect[(r, c)]
        label = val
        if val is None:
            label = ''
        self._svg.update_rect_element(gid, text=label)
        self._data[r][c] = val
    

    def shape(self):
        """
        Returns:
            tuple(int, int): Return the rows and columns of this table.
        """
        return (self._row, self._col)

    
    def reshape(self, row, col):
        """Reshape the row and column size of this table.
        row, col (int): The new row and column number of the table.

        Raises:
            AlgvizParamError: Table row or col number should > 0.
        """
        if row <=0 or col <=0:
            raise AlgvizParamError('Table row or col number should > 0.')
        if row > self._row:
            # Add new rows into table.
            for r in range(self._row, row):
                self._data.append(list())
                self._new_row_index_text_in_svg_(r)
                for c in range(col):
                    self._data[r].append(None)
                    self._new_rect_in_svg_(r, c)
        elif row < self._row:
            # Remove row in table.
            for r in range(row, self._row):
                self._data.pop()
                self._svg.delete_element(self._row_index2text[r])
                self._row_index2text.pop(r)
                for c in range(self._col):
                    self._svg.delete_element(self._index2rect[(r, c)])
                    self._index2rect.pop((r, c))
        if col > self._col:
            # Add new columns into table.
            for r in range(min(self._row, row)):
                for c in range(self._col, col):
                    self._data[r].append(None)
                    self._new_rect_in_svg_(r, c)
            for c in range(self._col, col):
                self._new_col_index_text_in_svg_(c)
        elif col < self._col:
            for r in range(min(self._row, row)):
                for c in range(col, self._col):
                    self._data[r].pop()
                    self._svg.delete_element(self._index2rect[(r, c)])
                    self._index2rect.pop((r, c))
            for c in range(col, self._col):
                self._svg.delete_element(self._col_index2text[c])
        self._row = row
        self._col = col
        self._update_svg_size_()
        self._update_subscripts_position_()


    def _add_cursor_(self, cursor, is_row):
        """Add a cursor into the cursor manager to track it.
        
        Args:
            cursor (Cursor): The cursor object to track.
        
        Returns:
            bool: Wheather add cursor successfully.
        """
        if type(cursor) != Cursor:
            return False
        need_update_svg = False
        if is_row:
            if not self._row_cursor_mgr.contains(cursor):
                self._row_cursor_mgr.add_cursor(cursor)
                self._col_cursor_mgr.on_svg_margin_changed((
                    self._cell_margin+self._row_cursor_mgr.get_cursors_occupy(), self._cell_margin))
                need_update_svg = True
        else:
            if not self._col_cursor_mgr.contains(cursor):
                self._col_cursor_mgr.add_cursor(cursor)
                self._row_cursor_mgr.on_svg_margin_changed((
                    self._cell_margin, self._cell_margin + self._col_cursor_mgr.get_cursors_occupy()))
                need_update_svg = True
        if need_update_svg:
            self._update_svg_size_()
            self._update_rects_position_()
            self._update_subscripts_position_()
            return True
        return False


    def _remove_cursor_(self, cursor):
        """Remove one cursor from Vector and it's SVG representation.

        Args:
            cursor (Cursor): The cursor object to be removed.
        
        Returns:
            bool: Wheather remove cursor successfully.
        """
        if type(cursor) != Cursor:
            return False
        need_update_svg = False
        if self._row_cursor_mgr.contains(cursor):
            self._row_cursor_mgr.remove_cursor(cursor)
            self._col_cursor_mgr.on_svg_margin_changed((
                self._cell_margin+self._row_cursor_mgr.get_cursors_occupy(), self._cell_margin))
            need_update_svg = True
        if self._col_cursor_mgr.contains(cursor):
            self._col_cursor_mgr.remove_cursor(cursor)
            self._row_cursor_mgr.on_svg_margin_changed((
                self._cell_margin, self._cell_margin+self._col_cursor_mgr.get_cursors_occupy()))
            need_update_svg = True
        if need_update_svg:
            self._update_svg_size_()
            self._update_rects_position_()
            self._update_subscripts_position_()
            return True
        return False


    def __getitem__(self, r):
        """
        Args:
            r (int/Cursor): The index of the row to access.
        
        Returns:
            TabRowIter: The iterator object for the row in the table.

        Raises:
            RuntimeError: Index:xx type is not int or Cursor.
            RuntimeError: Table index=xxx out of range.
        """
        return TableRowOperator(r, self)

    
    def __iter__(self):
        self._next_row = 0
        return self
    

    def __next__(self):
        if self._next_row >= self._row:
            raise StopIteration
        else:
            res = TableRowIter(self._next_row, self)
            self._next_row += 1
            return res
    

    def _repr_svg_(self):
        """
        Returns:
            str: The SVG representation of current table.
        """
        for (gid, color) in self._frame_trace_old:
            if (gid, color, True) not in self._frame_trace and (gid, color, False) not in self._frame_trace:
                self._cell_tcs[gid].remove(color)
            self._svg.update_rect_element(gid, fill=self._cell_tcs[gid].color())
        self._frame_trace_old.clear()
        for (gid, color, hold) in self._frame_trace:
            self._svg.update_rect_element(gid, fill=self._cell_tcs[gid].color())
            if not hold:
                self._frame_trace_old.append((gid, color))
        self._row_cursor_mgr.refresh_cursors_animation(self._row, (0, self._delay))
        self._col_cursor_mgr.refresh_cursors_animation(self._col, (0, self._delay))
        res_svg = self._svg._repr_svg_()
        self._frame_trace.clear()
        self._row_cursor_mgr.update_cursors_position()
        self._col_cursor_mgr.update_cursors_position()
        return res_svg


    def _update_svg_size_(self):
        svg_width = self._col*self._cell_width + self._cell_margin*2
        svg_height = self._row*self._cell_height + self._cell_margin*2
        svg_width += self._row_cursor_mgr.get_cursors_occupy()
        svg_height += self._col_cursor_mgr.get_cursors_occupy()
        if self._show_index:
            svg_width += len(str(self._row-1))*self._label_font_size
            svg_height += self._label_font_size
        self._svg.update_svg_size(svg_width, svg_height)


    def _update_rects_position_(self):
        for r in range(self._row):
            for c in range(self._col):
                rect_pos_x = c*self._cell_width+self._cell_margin+self._row_cursor_mgr.get_cursors_occupy()
                rect_pos_y = r*self._cell_height+self._cell_margin+self._col_cursor_mgr.get_cursors_occupy()
                rect = (rect_pos_x, rect_pos_y, self._cell_width, self._cell_height)
                gid = self._index2rect[(r, c)]
                self._svg.update_rect_element(gid, rect)


    def _update_subscripts_position_(self):
        for r in range(self._row):
            pos_x = self._col*self._cell_width+self._cell_margin*2+self._row_cursor_mgr.get_cursors_occupy()
            pos_y = (r+0.5)*self._cell_height+self._label_font_size*0.5+self._cell_margin+self._col_cursor_mgr.get_cursors_occupy()
            self._svg.update_text_element(self._row_index2text[r], (pos_x, pos_y))
        for c in range(self._col):
            pos_x = (c+0.5)*self._cell_width-self._label_font_size*len(str(c))*0.25+self._cell_margin+self._row_cursor_mgr.get_cursors_occupy()
            pos_y = self._row*self._cell_height+1+self._label_font_size+self._cell_margin+self._col_cursor_mgr.get_cursors_occupy()
            self._svg.update_text_element(self._col_index2text[c], (pos_x, pos_y))

    def _check_index_type_and_range_(self, index, max_val):
        res = None
        if type(index) is int:
            res = index
        elif type(index) is Cursor:
            res = index.index()
        else:
            raise RuntimeError('Index:{} type is not int or Cursor.'.format(index))
        if index < 0 or index >= max_val:
            raise RuntimeError('Table index={} out of range!'.format(index))
        return res

    def _new_rect_in_svg_(self, r, c):
        rect_pos_x = c*self._cell_width+self._cell_margin+self._row_cursor_mgr.get_cursors_occupy()
        rect_pos_y = r*self._cell_height+self._cell_margin+self._col_cursor_mgr.get_cursors_occupy()
        rect = (rect_pos_x, rect_pos_y, self._cell_width, self._cell_height)
        gid = self._svg.add_rect_element(rect, self._data[r][c], angle=False)
        self._index2rect[(r, c)] = gid
        self._cell_tcs[gid] = TraceColorStack()

    def _new_row_index_text_in_svg_(self, r):
        pos_x = self._col*self._cell_width+self._cell_margin*2+self._row_cursor_mgr.get_cursors_occupy()
        pos_y = (r+0.5)*self._cell_height+self._label_font_size*0.5+self._cell_margin+self._col_cursor_mgr.get_cursors_occupy()
        gid = self._svg.add_text_element((pos_x, pos_y), r, self._label_font_size)
        self._row_index2text[r] = gid

    def _new_col_index_text_in_svg_(self, c):
        pos_x = (c+0.5)*self._cell_width-self._label_font_size*len(str(c))*0.25+self._cell_margin+self._row_cursor_mgr.get_cursors_occupy()
        pos_y = self._row*self._cell_height+1+self._label_font_size+self._cell_margin+self._col_cursor_mgr.get_cursors_occupy()
        gid = self._svg.add_text_element((pos_x, pos_y), c, self._label_font_size)
        self._col_index2text[c] = gid



#!/usr/bin/env python3

"""Define the Logger related classes.

Author: zjl9959@gmail.com

License: GPLv3

"""

from algviz.utility import get_text_width

from xml.dom.minidom import Document


LOG_OFFSET_X = 5

class Logger():
    """A log class to display strings into the screen.

    You can write string into this logger object, it will separate lines by \\n.
    And if the buffer lines overflow, the oldest lines in the buffer will be discard.
    """

    def __init__(self, buffer_lines, font_size, show_line_num):
        """
        Args:
            buffer_lines (int): Set the max buffer lines for cached logs. Override the oldest line if overflow.
        """
        self._buffer_lines = max(buffer_lines, 1)
        font_size = int(font_size)
        self._font_size = 12
        if font_size > 1 and font_size <= 16:
            self._font_size = font_size
        self._logs = list()
        self._dom = Document()
        self._svg = self._dom.createElement('svg')
        self._svg.setAttribute('xmlns', 'http://www.w3.org/2000/svg')
        self._dom.appendChild(self._svg)
        self._next_node_id = 0
        self._log_lines = 0
        self._show_line_num = show_line_num
    

    def write(self, data):
        """Write log data. Use \\n to split multi-lines.
        
        Args:
            data (str): The log data string.
        """
        data_lines = data.split('\n')
        if len(data_lines) > self._buffer_lines:
            data_lines = data_lines[len(data_lines) - self._buffer_lines:]
        if len(data_lines) + len(self._logs) > self._buffer_lines:
            self._logs = self._logs[len(data_lines) + len(self._logs) - self._buffer_lines:]
        for line in data_lines:
            if self._show_line_num:
                self._logs.append('{}. {}'.format(self._log_lines, line))
            else:
                self._logs.append(line)
            self._log_lines += 1
    

    def clear(self):
        """Clear all cached log string.
        """
        for child in self._svg.childNodes:
            self._svg.removeChild(child)
        self._logs.clear()

    
    def _repr_svg_(self):
        svg_width = 0
        for child in self._svg.childNodes:
            self._svg.removeChild(child)
        g = self._dom.createElement('g')
        g.setAttribute('id', str(self._next_node_id))
        self._next_node_id += 1
        for i in range(len(self._logs)):
            log = self._logs[i]
            txt = self._dom.createElement('text')
            txt.setAttribute('x', '{}'.format(LOG_OFFSET_X))
            txt.setAttribute('y', '{:.2f}'.format(self._font_size*(i*1.2+1)))
            txt.setAttribute('font-size', '{:.2f}'.format(self._font_size))
            txt.setAttribute('font-family', 'cursive')
            text = self._dom.createTextNode('{}'.format(log))
            txt.appendChild(text)
            g.appendChild(txt)
            svg_width = max(svg_width, get_text_width(log, self._font_size) + LOG_OFFSET_X)
        self._svg.appendChild(g)
        # Update svg width and height.
        svg_height = len(self._logs)*self._font_size*1.3
        self._svg.setAttribute('width', '{:.0f}pt'.format(svg_width))
        self._svg.setAttribute('height', '{:.0f}pt'.format(svg_height))
        self._svg.setAttribute('viewBox', '0.00 0.00 {:.2f} {:.2f}'.format(svg_width, svg_height))
        return self._dom.toxml()



#!/usr/bin/env python3

"""Algviz is an algorithm visualization tool for your Python code.

Algviz can generate visual animations for
vector, table, linked list, tree, and graph data structures.
You can bring alive animations in your notebook after
inserting a few algviz interfaces into code. 

Author: zjl9959@gmail.com

License: GPLv3

"""

from .visual import Visualizer
from .graph import GraphNode, parseGraph, updateGraphEdge
from .tree import BinaryTreeNode, TreeNode, RecursiveTree, parseBinaryTree, parseTree
from .linked_list import ForwardLinkedListNode, DoublyLinkedListNode
from .linked_list import parseForwardLinkedList, parseDoublyLinkedList
from .utility import _version, AlgvizParamError, AlgvizRuntimeError, AlgvizFatalError, AlgvizTypeError


# Common colors name to RGB map (see: https://www.w3schools.com/tags/ref_colornames.asp)
color_aqua = (0, 255, 255)
color_black = (0, 0, 0)
color_blue = (0, 0, 255)
color_brown = (165, 42, 42)
color_dark_blue = (0, 0, 139)
color_dark_green = (0, 100, 0)
color_dark_red = (139, 0, 0)
color_gold = (255, 215, 0)
color_gray = (128, 128, 128)
color_green = (0, 128, 0)
color_green_yellow = (173, 255, 47)
color_light_blue = (173, 216, 230)
color_lime = (0, 255, 0)
color_orange = (255, 165, 0)
color_pink = (255, 192, 203)
color_plum = (221, 160, 221)
color_purple = (128, 0, 128)
color_red = (255, 0, 0)
color_silver = (192, 192, 192)
color_slate_blue = (106, 90, 205)
color_spring_green = (0, 255, 127)
color_tan = (210, 180, 140)
color_tomato = (255, 99, 71)
color_violet = (238, 130, 238)
color_wheat = (245, 222, 179)
color_white = (255, 255, 255)
color_yellow = (255, 255, 0)


__all__ = [
    'Visualizer',
    'GraphNode', 'parseGraph', 'updateGraphEdge',
    'BinaryTreeNode', 'TreeNode', 'RecursiveTree', 'parseBinaryTree', 'parseTree',
    'ForwardLinkedListNode', 'DoublyLinkedListNode',
    'parseForwardLinkedList', 'parseDoublyLinkedList',
    'AlgvizParamError', 'AlgvizRuntimeError', 'AlgvizFatalError', 'AlgvizTypeError'
]


__title__ = 'algviz'
__version__ = _version
__author__ = 'zjl9959@gmail.com'
__license__ = 'GNU General Public License (GPLv3)'
__help__ = "Visit https://algviz.com for more information."



#!/usr/bin/env python3

"""Define tree related classes and functions.

Author: zjl9959@gmail.com

License: GPLv3

"""

from algviz.utility import AlgvizParamError
from algviz.graph_node_base import GraphNodeBase as NodeBase


class BinaryTreeNode(NodeBase):
    """The definition of a binary tree node.

    Attributes:
        val (printable): The label to be displayed in the binary tree node.
        left (BinaryTreeNode): Point to the left subtree node object.
        right (BinaryTreeNode): Point to the right subtree node object.
    """

    def __init__(self, val, left=None, right=None):
        """
        Args:
            val (printable): The label value for binary node(should be printable object).
            left (BinaryTreeNode): Point to the left subtree node object.
            right (BinaryTreeNode): Point to the right subtree node object.
        """
        super().__init__(val)
        super().__setattr__('left', left)
        self._on_update_neighbor_(left)
        super().__setattr__('right', right)
        self._on_update_neighbor_(right)
    

    def __getattribute__(self, name):
        if name == 'left' or name == 'right':
            node = super().__getattribute__(name)
            return node
        else:
            return super().__getattribute__(name)
    

    def __setattr__(self, name, value):
        if name == 'left' or name == 'right':
            super().__setattr__(name, value)
            self._on_update_neighbor_(value)
        else:
            super().__setattr__(name, value)


    def _neighbors_(self):
        """Return all children nodes of this tree node.
        
        Returns:
            list(tuple(BinaryTreeNode, None)): Return left child node and right child node.
        """
        left_node = super().__getattribute__('left')
        right_node = super().__getattribute__('right')
        # Add edge label if this node just has single child node.
        left_label, right_label = None, None
        if left_node and not right_node:
            left_label = 'L'
        elif right_node and not left_node:
            right_label = 'R'
        return [(left_node, left_label), (right_node, right_label)]


def parseBinaryTree(tree_info):
    """ Create a new Tree from given node values.
    
    Args:
        tree_info (list(printable)): The label of each node in the tree must be given.
            Empty node is represented by None. eg:([1, None, 2, None, None, 3, 4])
    
    Returns:
        TreeNode: Root node object of this tree.
    """
    if len(tree_info) == 0:
        return None
    root = BinaryTreeNode(tree_info[0])
    node_queue = [root]
    index = 1
    while len(node_queue) > 0:
        cur_node = node_queue.pop(0)
        if index >= len(tree_info):
            break
        left_node = None
        if tree_info[index] is not None:
            left_node = BinaryTreeNode(tree_info[index])
        node_queue.append(left_node)
        if index + 1 >= len(tree_info):
            break
        right_node = None
        if tree_info[index + 1] is not None:
            right_node = BinaryTreeNode(tree_info[index+1])
        node_queue.append(right_node)
        if cur_node is not None:
            cur_node.left = left_node
            cur_node.right = right_node
        index += 2
    return root


class TreeChildrenIter():
    """Iterator for the children of TreeNode object.
    """
    def __init__(self, node, children):
        self._node = node
        self._children = children
        self._next_index = 0


    def __iter__(self):
        self._next_index = 0
        return self


    def __next__(self):
        if self._next_index >= len(self._children):
            raise StopIteration
        else:
            child = self._children[self._next_index]
            self._next_index += 1
            return child


class TreeNode(NodeBase):
    """A tree node has multiply children node.

    Attributes:
        val (printable): The label to be displayed in the binary tree node.
    """
    def __init__(self, val):
        """
        Args:
            val (printable): The label value for tree node.
        """
        super().__init__(val)
        super().__setattr__('_children', list())
    

    def children(self):
        """Return an iterator to iter over all the children nodes of this node.
        
        Returns:
            TreeChildrenIter: Children node iterator.
        """
        children_ = super().__getattribute__('_children')
        return TreeChildrenIter(self, tuple(children_))


    def childCount(self):
        """Return the children count of this node.

        Returns:
            int: children count.
        """
        children_ = super().__getattribute__('_children')
        return len(children_)


    def childAt(self, index):
        """Return the child node at the index position of the children list.
        
        Args:
            index (int): The position of the child node.

        Returns:
            TreeNode: child node.

        Raises:
            AlgvizParamError: TreeNode child index type error or out of range.
        """
        children_ = super().__getattribute__('_children')
        if type(index) != int or index < 0 or index >= len(children_):
            raise AlgvizParamError('TreeNode child index type error or out of range.')
        return children_[index]


    def childIndex(self, child):
        """Return the child node index in the children list.

        Args:
            child (TreeNode): The child node object to be located.

        Returns:
            int: the index position of the child node. If child not found, then return -1.
        """
        res = -1
        children_ = super().__getattribute__('_children')
        for i in range(len(children_)):
            if child == children_[i]:
                res = i
        return res


    def add(self, child, index=None):
        """Add a child node for this node.
        
        Args:
            child (TreeNode): The child node object to be added.
            index (int): The position to insert the child node.

        Raises:
            AlgvizParamError: TreeNode child index should be a positive integer!
        """
        if (type(index) != int and index != None) or (type(index) == int and index < 0):
            raise AlgvizParamError('TreeNode child index should be a positive integer!')
        children_ = super().__getattribute__('_children')
        for n in children_:
            if child == n:
                return
        if index == None or index >= len(children_):
            children_.append(child)
        else:
            children_.insert(index, child)
        self._on_update_neighbor_(child)


    def remove(self, child):
        """Remove one child node.
        
        Args:
            child (TreeNode): The child node to be removed.
        """
        children_ = super().__getattribute__('_children')
        for i in range(len(children_)):
            if child == children_[i]:
                children_.pop(i)
                self._on_update_neighbor_(None)
                return

    
    def removeAt(self, index):
        """Remove the child at the index position.

        Args:
            index (int): The child node index to be removed.

        Raises:
            AlgvizParamError: TreeNode child index type error or out of range.
        """
        children_ = super().__getattribute__('_children')
        if type(index) != int or index < 0 or index >= len(children_):
            raise AlgvizParamError('TreeNode child index type error or out of range.')
        children_.pop(index)
        self._on_update_neighbor_(None)


    def _neighbors_(self):
        children_ = super().__getattribute__('_children')
        res = list()
        for child in children_:
            res.append((child, None))
        return res


def parseTree(tree_info, nodes_label=None):
    """Create a Tree from node_map information and return the root node.
    
    Input tree_info should meet these constraints:
        
        1. Each node must have one parent node expect for the root node.

        2. The node id in tree info should be unique,
        but you can set the nodes_label to map the unique id into it's display label text.

    Args:
        tree_info (dict(printable:list(printable))): Describe the linked information of this tree.
            Key is the label of root node, Value is the list of it's children nodes label.
        nodes_label (dict(printable:printable)): Map the node id into it's display label.
    
    Returns:
        TreeNode: The root node of this tree.

    Raises:
        AlgvizParamError: (parseTree) node xxx have more than one parent node!
        AlgvizParamError: (parseTree) tree has more than one root node.
    """
    # Create TreeNode objects.
    nodes_dict = dict()
    child_nodes = set()     # Used to find the root node and check the validity of tree.
    for node, children in tree_info.items():
        # Create parent node.
        node_val = node
        if nodes_label and node in nodes_label:
            node_val = nodes_label[node]
        nodes_dict[node] = TreeNode(node_val)
        for child in children:
            # Create children nodes.
            if child not in child_nodes:
                child_nodes.add(child)
                node_val = child
                if nodes_label and child in nodes_label:
                    node_val = nodes_label[child]
                nodes_dict[child] = TreeNode(node_val)
            else:
                raise AlgvizParamError('(parseTree) node {} have more than one parent node!'.format(child))
    # Check and find the root node of this tree.
    root = None
    for node in tree_info.keys():
        if node not in child_nodes:
            if not root:
                root = node
            else:
                raise AlgvizParamError('(parseTree) tree has more than one({}, {}) root node.'.format(root, node))
    # Link the edge between root and it's child.
    for node, children in tree_info.items():
        parent_node = nodes_dict[node]
        for child in children:
            child_node = nodes_dict[child]
            parent_node.add(child_node)
    return nodes_dict[root]


class RecursiveTree():
    """Used to trace the recursive process in the recursive algorithm.
    """
    def __init__(self, viz, name="Recursive tree"):
        """
        Args:
            viz (Visualizer): The visualizer object that contains this recursive tree.
            name (printable): The display name of this recursive tree.
        """
        self.root = TreeNode("root")
        self.graph = viz.createGraph([self.root], name, True)
        self.stack = list()     # Recursive stack.

    def forward(self, val=''):
        """Forward to a new depth of this recursive tree.
        Args:
            val (printable): The label to be displayed in the recursive tree node.
        """
        node = TreeNode(val)
        self.root.add(node)
        self.graph.removeMark((173, 255, 47))
        self.graph.markNode((173, 255, 47), node, hold=True)
        self.stack.append(self.root)
        self.root = node

    def backward(self):
        """Backward to the last visited node in the recursive tree.
        """
        self.graph.markNode((192, 192, 192), self.root, hold=True)
        self.root = self.stack.pop()
        self.graph.removeMark((255, 99, 71))
        self.graph.markNode((255, 99, 71), self.root, hold=True)



#!/usr/bin/env python3

"""Define the base class for all the graph nodes.

Do not create GraphNodeBase class directly,
create one of it's subclass instead.

Author: zjl9959@gmail.com

License: GPLv3

"""

class GraphNodeBase:
    """Base class for all the graph nodes.
    
    GraphNodeBase's subclasses contain:
    GraphNode, BinaryTreeNode, ForwardLinkedListNode and DoublyLinkedListNode.
    If you want to get/set attribute in subclass, don't use dot operation like 'self.attr'.
    Please use these interfaces instead:

        super().__getattribute__(attr)
        
        super().__setattr__(attr, new_value)
    
    Attributes:
        val (printable): The label to be displayed in the graph node.
    """

    def __init__(self, val):
        """
        Args:
            val (printable): The initial label for the graph node.
        """
        object.__setattr__(self, 'val', val)
        object.__setattr__(self, '_bind_graphs', set())

    
    def __getattribute__(self, name):
        if name == 'val':
            return object.__getattribute__(self, 'val')
        else:
            return object.__getattribute__(self, name)


    def __setattr__(self, name, value):
        if name == 'val':
            object.__setattr__(self, 'val', value)
            self._on_update_value_(value)
        else:
            object.__setattr__(self, name, value)


    def __str__(self):
        return str(object.__getattribute__(self, 'val'))


    def _on_update_value_(self, value):
        """Notify the bind graph to update the displayed node mark and value.

        Args:
            value (printable): The value to be displayed in this node.
        """
        bind_graphs = object.__getattribute__(self, '_bind_graphs')
        for graph in bind_graphs:
            graph._updateNodeLabel(self, value)

    
    def _on_update_neighbor_(self, new_neighbor):
        """
        Notify the bind graph to update the displayed edge color between this node and it's neighbor.
        old_neighbor and new_neighbor can be None.

        Args:
            new_neighbor: The new neighbor node to replace the old neighbor nodes.
        """
        # Mark edge between this node and it's old_neighbor.
        bind_graphs = object.__getattribute__(self, '_bind_graphs')
        if len(bind_graphs) == 0:
            return
        # Mark edge between this node and it's new_neighbor.
        if new_neighbor:
            for graph in bind_graphs:
                graph.addNode(new_neighbor)


    def bind_graphs(self):
        """
        Returns:
            set(SvgGraph): The bind graphs collection of this node.
        """
        return object.__getattribute__(self, '_bind_graphs')


    def _bind_new_graph_(self, graph):
        """Bind a new SvgGraph object for this graph node object.
        
        Args:
            graph (SvgGraph): New SvgGraph object to track this node.
        """
        bind_graphs = object.__getattribute__(self, '_bind_graphs')
        if graph not in bind_graphs:
            bind_graphs.add(graph)


    def _remove_bind_graph_(self, graph):
        """Remove the SvgGraph object binded to this graph node object.
        
        The graph object will not track the change of this node, if it is successfully removed from bind_graphs.

        Args:
            graph (SvgGraph): SvgGraph object to be removed.
        """
        bind_graphs = object.__getattribute__(self, '_bind_graphs')
        if graph in bind_graphs:
            bind_graphs.remove(graph)



#!/usr/bin/env python3

"""Define low-level SVG animation refresh related classes for graph.

This module was used by linked_list, tree and graph module.
You can use the addNode, removeNode, markNode, markEdge and removeMark interfaces to update the graph.

Author: zjl9959@gmail.com

License: GPLv3

"""

from algviz.utility import str2rgbcolor, text_font_size, auto_text_color, rgbcolor2str
from algviz.utility import add_animate_appear_into_node, add_animate_move_into_node
from algviz.utility import TraceColorStack, ConsecutiveIdMap, AlgvizFatalError
from algviz.utility import add_desc_into_svg, find_tag_by_id
from algviz.graph import GraphNode
from algviz.tree import BinaryTreeNode, TreeNode
from algviz.linked_list import ForwardLinkedListNode, DoublyLinkedListNode

from graphviz import Digraph as graphviz_Digraph
from graphviz import Graph as graphviz_Graph
from graphviz import __version__ as graphviz_version
from xml.dom.minidom import parseString as mindom_parseString


class _SvgGraphType:
    """This class is used to specific the layout parameter for SvgGraph class.
    """
    def __init__(self, rankdir=None):
        """
        Args:
            rankdir (str): The layout direction for graph. example: 'LR'
        """
        self.rankdir = rankdir


def _get_graph_type_by_data_(data):
    """Check the input graph nodes data type and create graph layout parameter.
    
    Returns:
        _SvgGraphType: The layout data parmeter for SvgGraph.
    """
    if not data:
        return _SvgGraphType()
    # Get one node data from different type of data.
    if type(data)==dict:
        data = tuple(data.values())
    if (type(data)==list or type(data)==tuple) and len(data) > 0:
        data = data[0]
    layout = _SvgGraphType()
    # Check the data type and specific the layout parameter.
    if type(data) == GraphNode:
        # For normal graph layout.
        layout = _SvgGraphType('LR')
    elif type(data) == BinaryTreeNode:
        # For binary tree layout.
        layout = _SvgGraphType()
    elif type(data) == TreeNode:
        # For normal tree layout.
        layout = _SvgGraphType()
    elif type(data) == ForwardLinkedListNode:
        # For forward linked list layout.
        layout = _SvgGraphType('LR')
    elif type(data) == DoublyLinkedListNode:
        # For doubly linked list layout.
        layout = _SvgGraphType('LR')
    return layout


class SvgGraph():
    """A SvgGraph object can record all the nodes in it's binded graph.
    
    It will relayout the graph and generate animation when add/remove/replace node(s) in graph.
    It will also show the mark of graph node(s)&edge(s) visit status.
    Every time `__repr_svg__` function is called, SvgGraph object will return the latest svg string for this graph and prepare for a new frame.
    
    """
    
    def __init__(self, data, directed, delay):
        """
        Args:
            data (iterable): The root node(s) of the topology graph, used to initialize auxiliary data for this graph.
            directed (bool): Should this graph be directed graph or undirected.
            delay (float): Animation delay time between two animation frames.
        """
        self._directed = directed       # Whether the graph is a directed graph.
        self._delay = delay             # Delay time of each frame of animation.
        self._node_seq = list()         # The graph node(s) list arranged in a certain order.
        self._add_nodes = list()        # Record the externally added node(s) since last frame.
        self._remove_nodes = list()     # Record the externally deleted node(s) since last frame.
        self._edge_label = dict()       # Label information to be displayed on each edge of the graph.
        self._node_tcs = dict()         # Record the trajectory access information for all nodes in the current graph (node: ColorStack).
        self._edge_tcs = dict()         # Record the trajectory access information of all edges in the current graph ((start_node, end_node): ColorStack).
        self._node_appear = set()       # Record the collection of node(s) that appearing in the next frame of animation.
        self._node_disappear = set()    # Record the collection of node(s) that disappear in the next frame of animation.
        self._edge_appear = set()       # Record the collection of edge(s) that appearing in the next frame of animation.
        self._edge_disappear = set()    # Record the collection of edge(s) that disappear in the next frame of animation.
        self._node_move = set()         # Record the collection of nodes moving in the animation effect.
        self._frame_trace_old = list()  # Cache the node/edge related information that needs to be cleared in the previous frame (node_index/edge_index, ColorStack).
        self._frame_trace = list()      # Cache the node/edge related information to be refreshed in the next frame (node_index/edge_index, ColorStack, persistence).
        self._svg = None                # The svg object of the graph to be displayed.
        self._node_idmap = None         # Map node_index value to the corresponding node index in graphviz's output svg.
        self._edge_idmap = None         # Map edge_index value to the corresponding edge index in graphviz's output svg.
        self._add_history = set()       # Record all the nodes that have been added since graph created. Used to check duplicates when add/remove nodes in the graph.
        self._type = _get_graph_type_by_data_(data)
        # Init graph nodes and svg.
        (self._svg, self._node_idmap, self._edge_idmap) = self._create_svg_()
        self._init_graph_nodes(data)    # Traverse the data and add nodes into this graph.
        add_desc_into_svg(self._svg)


    def addNode(self, node):
        """Add a new node and all it's successor nodes into this graph.
        
        Args:
            node (subclass of GraphNodeBase): The node object to be added. Can be a graph/tree/linked_list node.
        
        Returns:
            int: The number of node(s) added into graph.
        """
        if node in self._add_history:
            return 0
        node_stack = [node]
        added_nodes_num = 0
        while len(node_stack) > 0:
            cur_node = node_stack.pop()
            if cur_node is None or cur_node in self._add_history:
                continue
            cur_node._bind_new_graph_(self)
            self._add_history.add(cur_node)
            self._add_nodes.append(cur_node)
            added_nodes_num = added_nodes_num + 1
            for neigh in cur_node._neighbors_()[::-1]:
                node_stack.append(neigh[0])
        return added_nodes_num
    

    def removeNode(self, node, recursive=False):
        """Remove a node from this graph. Remove this node's all successor nodes if recursive is True.
        
        If the removed node(s) have input edge(s) from the remains node(s) in this graph, do nothing a return 0.

        Args:
            node (subclass of GraphNodeBase): The node object to be removed. Can be a graph/tree/linked_list node.
            recursive (bool): Wheather to remove all the successor nodes of this node.
        
        Returns:
            int: The number of node(s) removed from graph.
        """
        subgraph_nodes = set()
        if recursive:
            node_stack = [node]
            while len(node_stack) > 0:
                cur_node = node_stack.pop()
                if cur_node is None or cur_node in subgraph_nodes:
                    continue
                if cur_node in self._remove_nodes or cur_node not in self._add_history:
                    continue
                cur_node._remove_bind_graph_(self)
                self._add_history.remove(cur_node)
                subgraph_nodes.add(cur_node)
                for neighbor in cur_node._neighbors_():
                    node_stack.append(neighbor[0])
        else:
            if node and node in self._add_history:
                subgraph_nodes.add(node)
            else:
                return 0
        # Make sure there is no output edge into the subgraph nodes to be removed.
        node_stack = self._node_seq + self._add_nodes
        visited = set(self._remove_nodes)
        while len(node_stack) > 0:
            cur_node = node_stack.pop()
            if cur_node is None or cur_node in subgraph_nodes or cur_node in visited:
                continue
            visited.add(cur_node)
            for neigh in cur_node._neighbors_():
                if neigh[0] is not None:
                    if neigh[0] in subgraph_nodes:
                        return 0
                    node_stack.append(neigh[0])
        for node in subgraph_nodes:
            self._remove_nodes.append(node)
        return len(subgraph_nodes)
    

    def markNode(self, color, node, hold=False):
        """Emphasize one node by mark it's background color.
        
        Args:
            color ((R,G,B)): The background color for the marked node. R, G, B stand for color channel for red, green, blue.
                R,G,B should be int value and 0 <= R,G,B <= 255. eg:(0, 255, 0)
            node (subclass of GraphNodeBase): The node object to be marked. Can be a graph/tree/linked_list node.
            hold (bool): Whether to keep the mark color in future animation frames.
        """
        if node is not None:
            if node not in self._node_tcs.keys():
                self._node_tcs[node] = TraceColorStack()
            self._node_tcs[node].add(color)
            self._frame_trace.append((node, color, hold))
    

    def markEdge(self, color, node1, node2, hold=False):
        """Emphasize one edge by mark it's stoke color.
        
        Args:
            color ((R,G,B)): The stroke color for the marked edge. R, G, B stand for color channel for red, green, blue.
                R,G,B should be int value and 0 <= R,G,B <= 255. eg:(0, 255, 0)
            node1, node2 (subclass of GraphNodeBase): The begin and end node in the edge to be marked. Can be a graph/tree/linked_list node.
            hold (bool): Whether to keep the mark color in future animation frames.
        """
        if node1 is not None and node2 is not None:
            edge_key = self._make_edge_tuple_(node1, node2)
            if edge_key not in self._edge_tcs.keys():
                self._edge_tcs[edge_key] = TraceColorStack(bgcolor=(123, 123, 123))
            self._edge_tcs[edge_key].add(color)
            self._frame_trace.append((edge_key, color, hold))
    

    def removeMark(self, color):
        """Remove the mark color for node(s) and edge(s).
        
        Args:
            color ((R,G,B)): R, G, B stand for color channel for red, green, blue.
                R,G,B should be int value and 0 <= R,G,B <= 255. eg:(0, 255, 0)
        """
        for k in self._node_seq:
            if self._node_tcs[k].remove(color):
                node_id = 'node{}'.format(self._node_idmap.toConsecutiveId(k))
                node = find_tag_by_id(self._svg, 'g', node_id)
                self._update_node_color_(node, self._node_tcs[k].color())
        for k in self._edge_label.keys():
            if self._edge_tcs[k].remove(color):
                edge_id = 'edge{}'.format(self._edge_idmap.toConsecutiveId(k))
                edge = find_tag_by_id(self._svg, 'g', edge_id)
                self._update_edge_color_(edge, self._edge_tcs[k].color())
    

    def _init_graph_nodes(self, data):
        if not data:
            return
        if type(data)==dict:
            for node in data.values():
                self.addNode(node)
        elif type(data)==list:
            for node in data:
                self.addNode(node)
        else:
            self.addNode(data)


    def _updateNodeLabel(self, node, label):
        """Update the label value of the node in the graph.
        
        Args:
            node (subclass of GraphNodeBase): The node object to be updated. Can be a graph/tree/linked_list node.
            label (printable): New label content.
        """
        node_id = 'node{}'.format(self._node_idmap.toConsecutiveId(node))
        svg_node = find_tag_by_id(self._svg, 'g', node_id)
        if svg_node is None or label is None:
            return
        ellipse = svg_node.getElementsByTagName('ellipse')[0]
        cx = float(ellipse.getAttribute('cx'))
        cy = float(ellipse.getAttribute('cy'))
        fc = str2rgbcolor(ellipse.getAttribute('fill'))
        text_svg = svg_node.getElementsByTagName('text')
        if text_svg is not None:
            svg_node.removeChild(text_svg[0])
        t = self._svg.createElement('text')
        t.setAttribute('alignment-baseline', 'middle')
        t.setAttribute('text-anchor', 'middle')
        t.setAttribute('font-family', 'Times,serif')
        t.setAttribute('x', '{:.2f}'.format(cx))
        t.setAttribute('y', '{:.2f}'.format(cy))
        font_size = min(14, text_font_size(32, '{}'.format(label)))
        t.setAttribute('font-size', '{:.2f}'.format(font_size))
        t.setAttribute('fill', auto_text_color(fc))
        tt = self._svg.createTextNode('{}'.format(label))
        t.appendChild(tt)
        svg_node.appendChild(t)
    

    def _updateEdgeLabel(self, node1, node2, label):
        """Update the label value of the node in the graph.
        
        Args:
            node (subclass of GraphNodeBase): The begin and end node in the edge to be updated. Can be a graph/tree/linked_list node.
            label (printable): New label content.
        """
        edge_key = self._make_edge_tuple_(node1, node2)
        edge_id = 'edge{}'.format(self._edge_idmap.toConsecutiveId(edge_key))
        svg_node = find_tag_by_id(self._svg, 'g', edge_id)
        if svg_node is None or label is None:
            return
        text_list = svg_node.getElementsByTagName('text')
        if text_list and len(text_list) > 0:
            text = text_list[0]
            for t in text.childNodes:
                text.removeChild(t)
            tt = self._svg.createTextNode('{}'.format(label))
            text.appendChild(tt)
    

    def _repr_svg_(self):
        """Render the graph into SVG and add animation effects.
        
        Returns:
            str: SVG string to representation graph nodes and edges with animation.
        """
        # Sequence the graph and add animation effects.
        self._traverse_graph_()
        (new_svg, node_idmap, edge_idmap) = self._create_svg_()
        add_desc_into_svg(new_svg)
        self._update_svg_size_(new_svg)
        self._update_svg_(new_svg, node_idmap, edge_idmap)
        self._update_trace_color_()
        self._compress_svg_()
        res = self._svg.toxml()
        # Update the SVG content and prepare for the next frame.
        self._svg, self._node_idmap, self._edge_idmap = new_svg, node_idmap, edge_idmap
        new_nodes = self._get_node_pos_(self._svg)
        for node_id in new_nodes.keys():
            node = self._node_idmap.toAttributeId(node_id)
            self._update_node_color_(new_nodes[node_id][0], self._node_tcs[node].color())
        new_edges = self._get_svg_edges_(self._svg)
        for edge_id in new_edges.keys():
            edge = self._edge_idmap.toAttributeId(edge_id)
            self._update_edge_color_(new_edges[edge_id], self._edge_tcs[edge].color())
        # Update auxiliary data caches.
        for node in self._node_disappear:
            self._node_tcs.pop(node)
        for edge in self._edge_disappear:
            self._edge_tcs.pop(edge)
        self._node_appear.clear()
        self._node_disappear.clear()
        self._edge_appear.clear()
        self._edge_disappear.clear()
        self._node_move.clear()
        return res.replace('\n', '')
    

    def _compress_svg_(self):
        graphs = self._svg.getElementsByTagName('g')
        for graph in graphs:
            if graph.getAttribute('class') != 'graph':
                continue
            nodes_to_remove = list()
            for node in graph.childNodes:
                if node.nodeType == node.COMMENT_NODE:
                    nodes_to_remove.append(node)
                elif node.nodeType == node.ELEMENT_NODE:
                    if node.getAttribute('class') == 'node':
                        titles = node.getElementsByTagName('title')
                        for title in titles:
                            node.removeChild(title)
                            pass
                    elif node.getAttribute('class') == 'edge':
                        titles = node.getElementsByTagName('title')
                        for title in titles:
                            title_content = title.firstChild.data
                            if self._directed:
                                title_content = title_content.split('->')
                            else:
                                title_content = title_content.split('--')
                            node.removeChild(title)
                    elif node.tagName == 'title':
                        nodes_to_remove.append(node)
            for node in nodes_to_remove:
                graph.removeChild(node)


    def _traverse_graph_(self):
        """Traverse each node in the graph and update the related data structure.
        """
        # Traverse the nodes in the graph and record the new topology structure of this graph.
        new_node_seq = list()
        new_edge_label = dict()
        visited = set(self._remove_nodes)
        node_stack = self._node_seq + self._add_nodes
        node_stack.reverse()
        while len(node_stack) > 0:
            cur_node = node_stack.pop()
            if cur_node is None or cur_node in visited:
                continue
            visited.add(cur_node)
            new_node_seq.append(cur_node)
            for neigh in cur_node._neighbors_()[::-1]:
                if neigh[0] is not None and neigh[0] not in self._remove_nodes:
                    temp_edge = self._make_edge_tuple_(cur_node, neigh[0])
                    new_edge_label[temp_edge] = neigh[1]
                    node_stack.append(neigh[0])
        # Update the newly added and disappeared edges and nodes.
        old_node_set = set(self._node_seq)
        new_node_set = set(new_node_seq)
        self._node_appear = new_node_set - old_node_set
        self._node_disappear = old_node_set - new_node_set
        self._edge_appear = new_edge_label.keys() - self._edge_label.keys()
        self._edge_disappear = self._edge_label.keys() - new_edge_label.keys()
        # Update the cached track color information.
        for node in self._node_appear:
            if node not in self._node_tcs.keys():
                self._node_tcs[node] = TraceColorStack()
        for edge in self._edge_appear:
            if edge not in self._edge_tcs.keys():
                self._edge_tcs[edge] = TraceColorStack(bgcolor=(123, 123, 123))
        # Update other auxiliary data.
        self._node_seq = new_node_seq
        self._edge_label = new_edge_label
        self._add_nodes.clear()
        self._remove_nodes.clear()
    

    def _update_node_color_(self, node, color):
        """Update the color attribute of the node in SVG.
        
        Args:
            node (xmldom.Node): The node to be updated in SVG.
            color ((R,G,B)): R, G, B stand for color channel for red, green, blue.
                R,G,B should be int value and 0 <= R,G,B <= 255. eg:(0, 255, 0)
        """
        if node is not None:
            ellipse = node.getElementsByTagName('ellipse')[0]
            ellipse.setAttribute('fill', rgbcolor2str(color))
            text_list = node.getElementsByTagName('text')
            if text_list and len(text_list) > 0:
                text = text_list[0]
                if text is not None:
                    text.setAttribute('fill', auto_text_color(color))
    

    def _update_edge_color_(self, edge, color):
        """Update the color attribute of the edge in SVG.
        
        Args:
            edge (xmldom.Node): The edge to be updated in SVG.
            color ((R,G,B)): R, G, B stand for color channel for red, green, blue.
                R,G,B should be int value and 0 <= R,G,B <= 255. eg:(0, 255, 0)
        """
        if edge is not None:
            path = edge.getElementsByTagName('path')[0]
            path.setAttribute('stroke', rgbcolor2str(color))
            polygons = edge.getElementsByTagName('polygon')
            if len(polygons) > 0:
                polygons[0].setAttribute('fill', rgbcolor2str(color))
                polygons[0].setAttribute('stroke', rgbcolor2str(color))
    

    def _update_trace_color_(self):
        """Update the color change of the SVG track in the frame.
        """
        for k, color in self._frame_trace_old:
            if type(k) == tuple and k in self._edge_tcs.keys():
                if (k, color, False) not in self._frame_trace and (k, color, True) not in self._frame_trace:
                    self._edge_tcs[k].remove(color)
                edge_id = 'edge{}'.format(self._edge_idmap.toConsecutiveId(k))
                edge = find_tag_by_id(self._svg, 'g', edge_id)
                self._update_edge_color_(edge, self._edge_tcs[k].color())
            elif k in self._node_tcs.keys():
                if (k, color, False) not in self._frame_trace and (k, color, True) not in self._frame_trace:
                    self._node_tcs[k].remove(color)
                node_id = 'node{}'.format(self._node_idmap.toConsecutiveId(k))
                node = find_tag_by_id(self._svg, 'g', node_id)
                self._update_node_color_(node, self._node_tcs[k].color())
        self._frame_trace_old.clear()
        for k, color, hold in self._frame_trace:
            if type(k) == tuple:
                edge_id = 'edge{}'.format(self._edge_idmap.toConsecutiveId(k))
                edge = find_tag_by_id(self._svg, 'g', edge_id)
                self._update_edge_color_(edge, self._edge_tcs[k].color())
            else:
                node_id = 'node{}'.format(self._node_idmap.toConsecutiveId(k))
                node = find_tag_by_id(self._svg, 'g', node_id)
                self._update_node_color_(node, self._node_tcs[k].color())
            if not hold:
                self._frame_trace_old.append((k, color))
        self._frame_trace.clear()
    

    def _update_svg_size_(self, new_svg):
        """Adjust the view size of self._svg to ensure that all elements can be observed.
        
        Args:
            new_svg (xmldom.Document): The latest SVG object to be updated.
        """
        old_svg_node = self._svg.getElementsByTagName('svg')[0]
        new_svg_node = new_svg.getElementsByTagName('svg')[0]
        old_svg_width = int(old_svg_node.getAttribute('width')[0:-2])
        old_svg_height = int(old_svg_node.getAttribute('height')[0:-2])
        new_svg_width = int(new_svg_node.getAttribute('width')[0:-2])
        new_svg_height = int(new_svg_node.getAttribute('height')[0:-2])
        width = max(old_svg_width, new_svg_width)
        height = max(old_svg_height, new_svg_height)
        old_svg_node.setAttribute('width', '{}pt'.format(width))
        old_svg_node.setAttribute('height', '{}pt'.format(height))
        old_svg_node.setAttribute('viewBox', '0.00 0.00 {:.2f} {:.2f}'.format(width, height))
        graph = find_tag_by_id(new_svg, 'g', 'graph0')
        clone_graph = graph.cloneNode(deep=False)
        clone_graph.setAttribute('id', 'graph1')
        old_svg_node.appendChild(clone_graph)
    

    def _update_svg_(self, new_svg, node_idmap, edge_idmap):
        """Add all node and edge related animations into SVG.
        
        Args:
            new_svg (xmldom.Doucment): The latest SVG object to be updated.
            node_idmap (ConsecutiveIdMap): The two-way mapping relationship between the node ID in the memory and the ID in the SVG.
            edge_idmap (ConsecutiveIdMap): The two-way mapping relationship between the edge ID in the memory and the ID in the SVG.
        """
        old_pos = self._get_node_pos_(self._svg)
        new_pos = self._get_node_pos_(new_svg)
        old_edges = self._get_svg_edges_(self._svg)
        new_edges = self._get_svg_edges_(new_svg)
        has_disappear_animate = False
        disappear_animate_end_time = self._delay*0.2
        # Add the disappearing animation effect for the graph nodes.
        for old_node_id in old_pos.keys():
            old_node = self._node_idmap.toAttributeId(old_node_id)
            if old_node in self._node_disappear:
                g = old_pos[old_node_id][0]
                animate = self._svg.createElement('animate')
                add_animate_appear_into_node(g, animate, (0, disappear_animate_end_time), False)
                has_disappear_animate = True
        # Update self._node_move because we need to use it when add disappearing animation effor for edges.
        for old_node_id in old_pos.keys():
            old_node = self._node_idmap.toAttributeId(old_node_id)
            if old_node in node_idmap._attr2id.keys():
                new_node_id = node_idmap.toConsecutiveId(old_node)
                delt_x = new_pos[new_node_id][1] - old_pos[old_node_id][1]
                delt_y = new_pos[new_node_id][2] - old_pos[old_node_id][2]
                if abs(delt_x) > 0.001 or abs(delt_y) > 0.001:  # Minimum delt move check.
                    self._node_move.add(old_node)
        # Add the disappearing animation effect for the edge.
        for old_edge_id in old_edges.keys():
            (node1, node2) = self._edge_idmap.toAttributeId(old_edge_id)  # The ID value of the edge (start_node, end_node) in the memory.
            if (node1, node2) in self._edge_disappear or node1 in self._node_move or node2 in self._node_move:
                g = old_edges[old_edge_id]
                animate = self._svg.createElement('animate')
                add_animate_appear_into_node(g, animate, (0, disappear_animate_end_time), False)
                has_disappear_animate = True
        # Add the moving animation effect for the graph nodes.
        has_move_animate = False
        move_animate_start_time = disappear_animate_end_time if has_disappear_animate else 0
        move_animate_end_time = move_animate_start_time + (self._delay-move_animate_start_time)*0.6
        for old_node_id in old_pos.keys():
            old_node = self._node_idmap.toAttributeId(old_node_id)
            if old_node in node_idmap._attr2id.keys():
                new_node_id = node_idmap.toConsecutiveId(old_node)
                delt_x = new_pos[new_node_id][1] - old_pos[old_node_id][1]
                delt_y = new_pos[new_node_id][2] - old_pos[old_node_id][2]
                if abs(delt_x) > 0.001 or abs(delt_y) > 0.001:  # Minimum delt move check.
                    g = old_pos[old_node_id][0]
                    animate = self._svg.createElement('animateMotion')
                    move = (delt_x, delt_y)
                    add_animate_move_into_node(g, animate, move, (move_animate_start_time, move_animate_end_time), False)
                    has_move_animate = True
        # Add the appearing animation effect for the graph nodes.
        graph = find_tag_by_id(self._svg, 'g', 'graph1')
        appear_animate_start_time = move_animate_end_time if has_move_animate else move_animate_start_time
        for old_node in self._node_appear:
            new_node_id = node_idmap.toConsecutiveId(old_node)
            old_node_id = self._node_idmap.toConsecutiveId(old_node)
            clone_node = new_pos[new_node_id][0].cloneNode(deep=True)
            clone_node.setAttribute('id', 'node{}'.format(old_node_id))
            graph.appendChild(clone_node)
            animate = self._svg.createElement('animate')
            add_animate_appear_into_node(clone_node, animate, (appear_animate_start_time, self._delay), True)
        # Add the appearing animation effect for the graph edges.
        graph = find_tag_by_id(self._svg, 'g', 'graph1')
        for new_edge_id in new_edges.keys():
            (node1, node2) = edge_idmap.toAttributeId(new_edge_id)
            if (node1, node2) in self._edge_appear or node1 in self._node_move or node2 in self._node_move:
                old_edge_id = self._edge_idmap.toConsecutiveId((node1, node2))
                clone_edge = new_edges[new_edge_id].cloneNode(deep=True)
                clone_edge.setAttribute('id', 'edge{}'.format(old_edge_id))
                graph.appendChild(clone_edge)
                animate = self._svg.createElement('animate')
                add_animate_appear_into_node(clone_edge, animate, (appear_animate_start_time, self._delay), True)
    

    def _get_node_pos_(self, svg):
        """Get the absolute coordinates of all the graph node(s) in the SVG.
        
        Args:
            svg (xmldom.Document): The SVG object to be display.
        
        Returns:
            dict(int:tuple(xmldom.Node,float,float)): The map from SVG_node_id to the SVG_node_object and node position.
                Key is SVG_node_id, value is (SVG_node_object, position_x, position_y).
        """
        graph = find_tag_by_id(svg, 'g', 'graph0')
        transform = graph.getAttribute('transform')
        translate_index = transform.find('translate')
        delt_x, delt_y = 0, 0
        if translate_index != -1:
            st = transform.find('(', translate_index) + 1
            ed = transform.find(')', translate_index)
            translate = transform[st:ed].split(' ')
            delt_x, delt_y = float(translate[0]), float(translate[1])
        positions = dict()
        nodes = svg.getElementsByTagName('g')
        for node in nodes:
            if node.getAttribute('class') == 'node':
                node_id = int(node.getAttribute('id')[4:])
                ellipse = node.getElementsByTagName('ellipse')[0]
                cx = float(ellipse.getAttribute('cx')) + delt_x
                cy = float(ellipse.getAttribute('cy')) + delt_y
                positions[node_id] = (node, cx, cy)
        return positions
    

    def _get_svg_edges_(self, svg):
        """Get all the edges index and object in the SVG xml tree.
        
        Args:
            svg (xmldom.Document): The SVG object to be display.
        
        Returns:
            dict(int:xmldom.Node): Key is the edge index in SVG, Value is the edge node in SVG.
        """
        edges = dict()
        nodes = svg.getElementsByTagName('g')
        for node in nodes:
            if node.getAttribute('class') == 'edge':
                edge_id = int(node.getAttribute('id')[4:])
                edges[edge_id] = node
        return edges
    

    def _make_edge_tuple_(self, node1, node2):
        """Create an edge tuple according to nodes and the graph's type.
        
        Args:
            node1, node2 (subclass of GraphNodeBase): The begin and end node in the edge. Can be a graph/tree/linked_list node.
        """
        if self._directed:
            return (node1, node2)
        else:
            if id(node1) < id(node2):
                return (node1, node2)
            else:
                return (node2, node1)
    

    def _create_svg_(self):
        """Call graphviz lib to create a new SVG object to represent the latest graph.
        The SVG is static and don't include animations.
        
        Returns:
            xmldom.Document: The xmldom.Document object of the SVG.

        Raises:
            AlgvizFatalError: Unsupported graphviz version xxx.
        """
        dot = None
        node_idmap = ConsecutiveIdMap(1)
        edge_idmap = ConsecutiveIdMap(1)
        if self._directed:
            dot = graphviz_Digraph(format='svg')
        else:
            dot = graphviz_Graph(format='svg')
        if self._type.rankdir == 'LR':
            dot.graph_attr['rankdir'] = 'LR'
        dot.graph_attr['bgcolor'] = '#00000000'
        dot.node_attr.update(shape='circle', fixedsize='shape', color='#7B7B7B')
        dot.edge_attr.update(arrowhead='vee', color='#7B7B7B')
        for node in self._node_seq:
            node_id = node_idmap.toConsecutiveId(node)
            if node is None:
                dot.node(name='{}'.format(node_id))
            else:
                fs = min(14, text_font_size(32, str(node)))
                dot.node(name='{}'.format(node_id), label='{}'.format(str(node)), fontsize='{:.2f}'.format(fs))
        for (node1, node2) in self._edge_label.keys():
            label = self._edge_label[(node1, node2)]
            node1_id = node_idmap.toConsecutiveId(node1)
            node2_id = node_idmap.toConsecutiveId(node2)
            if label is None:
                dot.edge('{}'.format(node1_id), '{}'.format(node2_id))
            else:
                dot.edge('{}'.format(node1_id), '{}'.format(node2_id), label='{}'.format(label), fontcolor='#C0C0C0', fontsize='12')
            edge_idmap.toConsecutiveId((node1, node2))
        raw_svg_str = ''
        try:
            callable(getattr(dot, '_repr_svg_'))
            raw_svg_str = dot._repr_svg_()
        except:
            try:
                # graphviz replaced interface '_repr_svg_' since version 0.19 (https://graphviz.readthedocs.io/en/stable/changelog.html#version-0-19)
                callable(getattr(dot, '_repr_image_svg_xml'))
                raw_svg_str = dot._repr_image_svg_xml()
            except:
                raise AlgvizFatalError('Unsupported graphviz version {}'.format(graphviz_version))
        return (mindom_parseString(raw_svg_str), node_idmap, edge_idmap)



#!/usr/bin/env python3

"""Define linked list related classes and functions.

Including ForwardLinkedListNode, DoublyLinkedListNode class
and parseForwardLinkedList, parseDoublyLinkedList function to parse linked list from string.

Typical usage example:

  head = parseForwardLinkedList([1, 2, 3])  # Create a forward linked list: 1->2-3
  
  new_head = ForwardLinkedListNode(0)       # Create a forward linked list node: 0
  
  new_head.next = head                      # Link node into linked list, and get: 0->1->2->3

Author: zjl9959@gmail.com

License: GPLv3

"""

from algviz.graph_node_base import GraphNodeBase as NodeBase


class ForwardLinkedListNode(NodeBase):
    """The node for forward linked list.

    Attributes:
        val (printable): The label to be displayed in the forward linked list node.
        next (ForwardLinkedListNode): Point to the next ForwardLinkedListNode object.
    """

    def __init__(self, val, next=None):
        """
        Args:
            val (printable): The label to be displayed in forward linked list node.
            next (ForwardLinkedListNode): Point to the next ForwardLinkedListNode object.
        """
        super().__init__(val)
        super().__setattr__('next', next)
        self._on_update_neighbor_(next)


    def __getattribute__(self, name):
        if name == 'next':
            # Visit next node in the forward linked list and update the node's visit state.
            node = super().__getattribute__('next')
            return node
        else:
            return super().__getattribute__(name)
    

    def __setattr__(self, name, value):
        if name == 'next':
            super().__setattr__('next', value)
            self._on_update_neighbor_(value)
        else:
            super().__setattr__(name, value)
    

    def _neighbors_(self):
        """Interface for SvgGraph. Get this node's neighbors(next node).

        Returns:
            list(tuple(next_node, None)): The neighbors list of this node.
        """
        next_node = super().__getattribute__('next')
        return [(next_node, None)]


class DoublyLinkedListNode(NodeBase):
    """The node for doubly linked list.

    Attributes:
        val (printable): The label to be displayed in the doubly linked list node.
        prev (DoublyLinkedListNode): Point to the previous DoublyLinkedListNode object.
        next (DoublyLinkedListNode): Point to the next DoublyLinkedListNode object.
    """

    def __init__(self, val, prev=None, next=None):
        """
        Args:
            val (printable): The label to be displayed in doubly linked list node.
            prev (DoublyLinkedListNode): Point to the previous DoublyLinkedListNode object.
            next (DoublyLinkedListNode): Point to the next DoublyLinkedListNode object.
        """
        super().__init__(val)
        super().__setattr__('prev', prev)
        self._on_update_neighbor_(prev)
        super().__setattr__('next', next)
        self._on_update_neighbor_(next)


    def __getattribute__(self, name):
        if name == 'next' or name == 'prev':
            # Visit next node in the doubly linked list and update the node's visit state.
            node = super().__getattribute__(name)
            return node
        else:
            return super().__getattribute__(name)


    def __setattr__(self, name, value):
        if name == 'next' or name == 'prev':
            super().__setattr__(name, value)
            self._on_update_neighbor_(value)
        else:
            super().__setattr__(name, value)


    def _neighbors_(self):
        """Interface for SvgGraph. Get this node's neighbors(next node).
        
        Returns:
            list(tuple(next_node, None)): The neighbors list of this node.
        """
        next_node = super().__getattribute__('next')
        prev_node = super().__getattribute__('prev')
        return [(next_node, None), (prev_node, None)]


def parseForwardLinkedList(list_info):
    """Create a new forward linked list object and return it's head node.
    
    Args:
        list_info (list(printable)): The labels to display in the forward linked list's nodes.
    
    Returns:
        ForwardLinkedListNode: The head node objet for this forward linked list.
    """
    if len(list_info) == 0:
        return None
    head = ForwardLinkedListNode(list_info[0])
    cur_node = head
    for i in range(1, len(list_info)):
        cur_node.next = ForwardLinkedListNode(list_info[i])
        cur_node = cur_node.next
    return head


def parseDoublyLinkedList(list_info):
    """Create a new doubly linked list object and return it's head and tail node.
    
    Args:
        list_info (list(printable)): The labels to display in the doubly linked list's nodes.
    
    Returns:
        DoublyLinkedListNode, DoublyLinkedListNode: The head and tail node objects for this doubly linked list.
    """
    if len(list_info) == 0:
        return (None, None)
    head = DoublyLinkedListNode(list_info[0])
    cur_node, next_node = head, head
    for i in range(1, len(list_info)):
        next_node = DoublyLinkedListNode(list_info[i])
        cur_node.next = next_node
        next_node.prev = cur_node
        cur_node = cur_node.next
    return head, next_node



#!/usr/bin/env python3

"""Define visual class in jupyter notebook.

Author: zjl9959@gmail.com

License: GPLv3

"""

from weakref import WeakKeyDictionary
from time import sleep
from IPython import display

from algviz.table import Table
from algviz.vector import Vector
from algviz.svg_graph import SvgGraph
from algviz.svg_table import SvgTable
from algviz.logger import Logger
from algviz.cursor import Cursor, _CursorRange
from algviz.utility import AlgvizParamError, AlgvizTypeError, kMaxNameChars


class _NoDisplay():
    def _repr_svg_(self):
        return ''


class _NameDisplay():
    def __init__(self, name):
        if type(name) != str or len(name) > kMaxNameChars:
            raise AlgvizParamError('name "{}" not str or exceed {} characters.'.format(name, kMaxNameChars))
        self._name = name + ':'

    def __repr__(self):
        return self._name


_next_display_id = 0


class Visualizer(): 

    def __init__(self, delay=2.0, wait=0.5):
        """
        Args:
            delay (float): Animation delay time (in seconds).
            wait (True/float/int): (True) wait for the key input to continue execute the code.
                                   (float/int) the wait time before start the next frame of animation.
        """
        self._delay = 2.0               # Set default delay time for animation as 3.0 seconds.
        if delay > 0:
            self._delay = delay
        self._wait = wait
        if type(self._wait) != bool and type(self._wait) != float and type(self._wait) != int:
            self._wait = 0.5
        if type(self._wait) == bool and self._wait == False:
            self._wait = 0.5
        if type(self._wait) != bool and self._wait < 0:
            self._wait = 0
        
        # The mapping relationship between the display object and the display id.
        self._element2display = WeakKeyDictionary()
        # Record the displayed id. If object is not displayed, call the display interface, otherwise call the update interface.
        self._displayed = set()
        # The mapping relationship between the display object and the object's name.
        self._displayid2name = dict()
        # The next unique cursor id created by this visualizer.
        self._next_cursor_id = -1


    def display(self, delay=None):
        """Refresh all created display objects.
        
        Args:
            delay (float): Specific the delay time for this animation frame (in seconds).
        """
        if delay == None:
            delay = self._delay
        if type(self._wait) == float or type(self._wait) == int:
            for elem in self._element2display.keyrefs():
                did = self._element2display[elem()]
                if did not in self._displayed:
                    if did in self._displayid2name:
                        svg_title = _NameDisplay(self._displayid2name[did])
                        display.display(svg_title, display_id='algviz_{}'.format(did))
                    elem()._delay = delay
                    display.display(elem(), display_id='algviz{}'.format(did))
                    self._displayed.add(did)
                else:
                    if did in self._displayid2name:
                        svg_title = _NameDisplay(self._displayid2name[did])
                        display.update_display(svg_title, display_id='algviz_{}'.format(did))
                    elem()._delay = delay
                    display.update_display(elem(), display_id='algviz{}'.format(did))
            temp_displayed = list(self._displayed)
            for did in temp_displayed:
                if did not in self._element2display.values():
                    if did in self._displayid2name:
                        display.update_display(_NoDisplay(), display_id='algviz_{}'.format(did))
                    display.update_display(_NoDisplay(), display_id='algviz{}'.format(did))
                    self._displayed.remove(did)
            sleep(delay + self._wait)
            return None
        else:
            display.clear_output(wait=True)
            for elem in self._element2display.keyrefs():
                did = self._element2display[elem()]
                if did in self._displayid2name:
                    svg_title = SvgTable(400, 17)
                    title_name = '{}:'.format(self._displayid2name[did])
                    svg_title.add_text_element((4, 14), title_name, font_size=14, fill=(0,0,0))
                    display.display(svg_title, display_id='algviz_{}'.format(did))
                elem()._delay = delay
                display.display(elem(), display_id='algviz{}'.format(did))
                self._displayed.add(did)
            return input('Input `Enter` to continue:')


    def createTable(self, row, col, data=None, name=None, cell_size=(40, 40), show_index=True):
        """
        Args:
            row, col (int): The number of rows, columns for this table.
            data (list(list(printable))): The initial data for table cells.
            name (str): The name of this table object.
            cell_size tuple(float, float): Table cell size (width, height).
            show_index (bool): Whether to display table row and column labels.
        
        Returns:
            Table: New created Table object.
        """
        global _next_display_id
        tab = Table(row, col, data, cell_size, show_index)
        self._element2display[tab] = _next_display_id
        if name is not None:
            self._displayid2name[_next_display_id] = name
        _next_display_id += 1
        return tab


    def createVector(self, data=None, name=None, cell_size=(40, 40), histogram=False, show_index=True):
        """
        Args:
            data (list(printable)): The initial data for vector cells.
            name (str): The name of this Vector object.
            cell_size tuple(float, float): Vector cell size (width, height).
            histogram (bool): Display the data in the form of a histogram or not.
            show_index (bool): Whether to display the vector index label.
        
        Returns:
            Vector: New created Vector object.
        """
        global _next_display_id
        vec = Vector(data, self._delay, cell_size, histogram, show_index)
        self._element2display[vec] = _next_display_id
        if name is not None:
            self._displayid2name[_next_display_id] = name
        _next_display_id += 1
        return vec


    def createGraph(self, data=None, name=None, directed=True):
        """
        Args:
            data (iterable): The root node(s) to initialize the topology graph.
            name (str): The name of this Vector object.
            directed (bool): Should this graph be directed graph or undirected.
        
        Returns:
            SvgGraph: Created SvgGraph object.
        """
        global _next_display_id
        gra = SvgGraph(data, directed, self._delay)
        self._element2display[gra] = _next_display_id
        if name is not None:
            self._displayid2name[_next_display_id] = name
        _next_display_id += 1
        return gra


    def createLogger(self, buffer_lines=10, name=None, font_size=12, show_line_num=True):
        """
        Args:
            buffer_lines (int): Maximum buffer line of this logger.
            name (str): The name of this Vector object.
        
        Returns:
            Logger: Created Logger object.
        """
        global _next_display_id
        logg = Logger(buffer_lines, font_size, show_line_num)
        self._element2display[logg] = _next_display_id
        if name is not None:
            self._displayid2name[_next_display_id] = name
        _next_display_id += 1
        return logg


    def createCursor(self, offset=0, name=None):
        """
        Args:
            offset (int/Cursor): The cursor's index position offset.
            name (str): The name of this Cursor object.

        Returns:
            Cursor: Created Cursor object.
        """
        if type(offset) is Cursor:
            offset = offset.index()
        self._next_cursor_id += 1
        return Cursor(name, offset, self._next_cursor_id)


    def removeCursor(self, cursor):
        """
        Args:
            cursor (Cursor): The cursor object to be removed from this visualizer.
        """
        if type(cursor) != Cursor:
            raise AlgvizTypeError('removeCursor object is not a Cursor type.')
        for elem in self._element2display.keyrefs():
            element = elem()
            if not element or (type(element) != Vector and type(element) != Table):
                continue
            element._remove_cursor_(cursor)


    def cursorRange(self, st, ed, step=1, name=None):
        """
        Args:
            st/ed (int/Cursor): The start/end of cursor index.
            step (int/Cursor): The change step of this cursor's index(default->1).
            name (str): The display name of this cursor(default->None).

        Returns:
            _CursorRange: the iterator of Cursor objects.
        """
        if type(st) is Cursor:
            st = st.index()
        if type(ed) is Cursor:
            ed = ed.index()
        self._next_cursor_id += 1
        return _CursorRange(self._next_cursor_id, name, st, ed, step)



#!/usr/bin/env python3

"""Some utility functions and classes used by other modules.

Author: zjl9959@gmail.com

License: GPLv3

"""

_version = '0.1.8'                  # algviz version
_url = 'https://algviz.com'         # The homepage for algviz
KFATAL_HELP_INFO = """You can report this bug from link: https://github.com/zjl9959/algviz/issues"""


# Define exceptions for algviz runtime.
class AlgvizParamError(Exception):
    def __init__(self, message):
        super().__init__('[AlgvizParamError] {}'.format(message))


class AlgvizRuntimeError(Exception):
    def __init__(self, message):
        super().__init__('[AlgvizRuntimeError] {}\n{}'.format(message, KFATAL_HELP_INFO))


class AlgvizFatalError(Exception):
    def __init__(self, message):
        super().__init__('[AlgvizFatalError] {}\n{}'.format(message, KFATAL_HELP_INFO))


class AlgvizTypeError(Exception):
    def __init__(self, obj):
        super().__init__('[AlgvizTypeError] Type {} not support!\n{}'.format(type(obj), KFATAL_HELP_INFO))


# Parameter limit for display data objects.
kMinAnimDelay = 0.1         # The minimum animation delay time is 0.1 seconds.
kMaxAnimDelay = 100.0       # The maximum animation delay time is 100.0 seconds.
kMinCellWidth = 10.0        # The minimum vector/table cell display width.
kMaxCellWidth = 100.0       # The maximum vector/table cell display width.
kMinCellHeight = 10.0       # The minimum vector/table cell display height.
kMaxCellHeight = 100.0      # The maximum vector/table cell display height.
kMaxBarHight = 1000.0       # The maximum histogram bar height of vector.
kMaxNameChars = 50          # The Maximum name character numbers.


class TraceColorStack():
    """Manage multiple colors on an element, perform color fusion operations.
    """

    def __init__(self, bgcolor=(255, 255, 255)):
        """
        Args:
            bgcolor ((R,G,B)): R, G, B stand for color channel for red, green, blue.
                R,G,B should be int value and 0 <= R,G,B <= 255. eg:(0, 255, 0)
        """
        self._colors = list()
        self._bgcolor = bgcolor
    

    def add(self, color):
        """Add a new color into TraceColorStack.
        
        Args:
            color ((R,G,B)): R, G, B stand for color channel for red, green, blue.
                R,G,B should be int value and 0 <= R,G,B <= 255. eg:(0, 255, 0)
        """
        if not len(self._colors) or color != self._colors[-1]:
            self._colors.append(color)
    

    def remove(self, color):
        """Remove color from TraceColorStack.
        
        Args:
            color ((R,G,B)): R, G, B stand for color channel for red, green, blue.
                R,G,B should be int value and 0 <= R,G,B <= 255. eg:(0, 255, 0)
        
        Returns:
            bool: Return False if can't color. Return True if successfully deleted color.
        """
        colors_new = list()
        for i in range(len(self._colors)):
            if color != self._colors[i]:
                colors_new.append(self._colors[i])
        res = False
        if len(colors_new) < len(self._colors):
            res = True
        self._colors = colors_new
        return res
    

    def color(self):
        """Get the merged color in TraceColorStack.
        
        Returns:
            color ((R,G,B)): R, G, B stand for color channel for red, green, blue.
        """
        if len(self._colors) == 0:
            return self._bgcolor
        return self._colors[-1]


class ConsecutiveIdMap():
    """Allocate contiguous integer numbers for hashable objects.
    """

    def __init__(self, offset=0):
        """
        Args:
            offset (int): The initial value of the initial mapping ID.
        """
        self._offset = offset
        self._next_id = offset
        self._attr2id = dict()
        self._id2attr = list()
    

    def toConsecutiveId(self, attr_id):
        """Create or get the continuous ID for an unordered ID.
        
        Args:
            attr_id (hashable): Unordered ID object.
        
        Returns:
            int: Continuous ID value.
        """
        if attr_id in self._attr2id.keys():
            return self._attr2id[attr_id]
        else:
            self._attr2id[attr_id] = self._next_id
            self._id2attr.append(attr_id)
            self._next_id += 1
            return self._next_id - 1
    

    def toAttributeId(self, cons_id):
        """Given a continuous ID, return it's correspond unordered ID.
        
        Args:
            cons_id (int): Continuous ID value.
       
        Returns:
            hashable: Unordered ID object.
        """
        return self._id2attr[cons_id - self._offset]


# TODO: Deprecated this function, use a more effective way to find node.
# Or, just cache the node object directly.
def find_tag_by_id(node, tag_name, tag_id):
    """Find the first match node in XML node and its sub nodes.
    Args:
        node (xmldom.Node): The XML node object to search.
        tag_name (str): The tag name of the element.
        tag_id (str) The id value of the element.
    
    Returns:
        xmldom.Node or None: Return the XML node object if found it, otherwise return None.
    """
    tags = node.getElementsByTagName(tag_name)
    for tag in tags:
        if tag.getAttribute('id') == tag_id:
            return tag
    return None


def clear_svg_animates(svg):
    """Clear all the animation effects in SVG.
    Args:
        svg (xmldom.Node): The SVG object to be cleared.
    """
    gg = svg.getElementsByTagName('g')
    for g in gg:
        animates_appear = g.getElementsByTagName('animate')
        if len(animates_appear):
            g.removeAttribute('style')
        animates_move = g.getElementsByTagName('animateMotion')
        for animate in animates_appear + animates_move:
            g.removeChild(animate)


def add_animate_move_into_node(g, animate, move, time, bessel):
    """
    Args:
        g (xmldom.Node): The SVG node to add move animation into.
        animate (xmldom.Node): The animate node to be added into node g.
        move (tuple(float, float)): (delt_x, delt_y) The delt move distance along x axis and y axis for this element.
        time (tuple(float, float)): (begin, end) The begin and end time of this animation.
        bessel (bool): Whether to set the path of this move animation as bezier curve.
    """
    g.appendChild(animate)
    if bessel:
        animate.setAttribute('path', 'm0,0 q{:.2f},{:.2f} {:.2f},{:.2f}'.format(move[0]*0.5-move[1]*0.2, move[1]*0.5+move[0]*0.2, move[0], move[1]))
    else:
        animate.setAttribute('path', 'm0,0 l{:.2f},{:.2f}'.format(move[0], move[1]))
    animate.setAttribute('begin', '{:.2f}s'.format(time[0]))
    animate.setAttribute('dur', '{:.2f}s'.format(time[1]-time[0]))
    animate.setAttribute('fill', 'freeze')


def add_animate_appear_into_node(g, animate, time, appear=True):
    """
        g (xmldom.Node): The SVG node to add appear animation into.
        animate (xmldom.Node): The animate node to be added into node g.
        time (tuple(float, float)) (begin, end) The begin and end time of this animation.
        appear (bool): True for appear animation; False for disappear animation.
    """
    g.setAttribute('style', 'opacity:{:.0f}'.format(not appear))
    g.appendChild(animate)
    animate.setAttribute('attributeName', 'opacity')
    animate.setAttribute('from', '{:.0f}'.format(not appear))
    animate.setAttribute('to', '{:.0f}'.format(appear))
    animate.setAttribute('begin', '{:.2f}s'.format(time[0]))
    animate.setAttribute('dur', '{:.2f}s'.format(time[1]-time[0]))
    animate.setAttribute('fill', 'freeze')


def auto_text_color(back_color):
    """Auto pick one text stroke color according to it's background color.
    
    Args:
        back_color ((R,G,B)): Text background color. R, G, B stand for color channel for red, green, blue.
            R,G,B should be int value and 0 <= R,G,B <= 255. eg:(0, 255, 0)
    
    Returns:
        str: Text stroke color value formatted with hexadecimal number(SVG format).
            eg: '#FFFFFF'
    """
    rgb_sum = back_color[0]+back_color[1]+back_color[2]
    if rgb_sum < 150:
        return '#FFFFFF'
    else:
        return '#000000'


def rgbcolor2str(color):
    """Convert (R, G, B) formatted color into hexadecimal formatted string.
    
    Args:
        color ((R,G,B)): R, G, B stand for color channel for red, green, blue.
            R,G,B should be int value and 0 <= R,G,B <= 255. eg:(0, 255, 0)
    
    Returns:
        str: Hexadecimal formatted string. (SVG format). eg: '#FFFFFF'
    """
    return '#{:0>2x}{:0>2x}{:0>2x}'.format(color[0], color[1], color[2])


def str2rgbcolor(color_str):
    """Convert hexadecimal formatted string into (R, G, B) formatted color.
    
    Args:
        color_str (str): Hexadecimal formatted string. (SVG format). eg: '#FFFFFF'
    
    Returns:
        (R,G,B): R, G, B stand for color channel for red, green, blue.
            R,G,B should be int value and 0 <= R,G,B <= 255. eg:(0, 255, 0)
    """
    color_str = color_str.strip('#')
    return (int(color_str[0:2], 16), int(color_str[2:4], 16), int(color_str[4:6], 16))


def text_font_size(text_width, text):
    """Calculate the font size based on the total width of the text and the text content.
        
    Args:
        text_width (float): The total width of the text.
        text (str): The text content (should be unicode format string).
    
    Returns:
        float: Text font size.
    """
    display_len = text_char_num(text)
    if display_len > 0:
        return min(16, text_width*1.6/display_len, text_width*0.8)
    else:
        return 0


def get_text_width(text, font_size):
    """Calculate the total text width of the given text string.
    
    Args:
        font_size (int): The font size of text.
        text (str): The text content (should be unicode format string).
    
    Returns:
        float: The total width of the text.
    """
    # TODO: Use more specific method to measure the text total width.
    text_num = text_char_num(text)
    return text_num*font_size*0.5


def text_char_num(text):
    """Count the number of characters in the text.
    
    Args:
        text (str): The text content (should be unicode format string).
    
    Returns:
        int: The number of characters in the text.
    """
    count = 0
    for ch in text:
        if '\u4e00' <= ch <= '\u9fff':
            count += 2
        else:
            count += 1
    return count


def clamp(val, min_val, max_val):
    """
    Returns:
        The clamped value between min_val and max_val.
    """
    return max(min(val, max_val), min_val)


def add_desc_into_svg(dom):
    """Add description meta data into SVG dom tree.
    
    Args:
        dom (xmldom.document) The dom object to contain the description.
    """
    # Check if there is already a description in svg.
    svgs = dom.getElementsByTagName('svg')
    if len(svgs) == 0:
        return
    svg = svgs[0]
    descs = svg.getElementsByTagName('desc')
    if len(descs) > 0:
        return
    desc_str = 'Generated by algviz-{}(see {}).'.format(_version, _url)
    desc = dom.createElement('desc')
    text = dom.createTextNode('{}'.format(desc_str))
    desc.appendChild(text)
    svg.appendChild(desc)


def add_default_text_style(dom):
    """Add the default text style into svg.
    
    Args:
        dom (xmldom.document) The dom object to contain the description.
    """
    svgs = dom.getElementsByTagName('svg')
    if len(svgs) == 0:
        return
    svg = svgs[0]
    style = dom.createElement('style')
    style_content = r".txt { alignment-baseline:middle; text-anchor:middle; font-family:Times,serif; }"
    text = dom.createTextNode(style_content)
    style.appendChild(text)
    svg.appendChild(style)



#!/usr/bin/env python3

"""

Author: zjl9959@gmail.com

License: GPLv3

"""

from os import path
from setuptools import setup

script_path = path.split(path.realpath(__file__))[0]
readme_filepath = path.join(script_path, 'README.md')
long_description_data = ''
with open(readme_filepath, encoding='utf-8') as f:
    long_description_data = f.read()

setup(
    name="algviz",
    version="0.1.8",
    author="zjl9959",
    author_email="zjl9959@gmail.com",
    description=("An algorithm visualization tool for jupyter notebook to show animation for vector, table, linked list, tree and graph data structures."),
    long_description=long_description_data,
    long_description_content_type="text/markdown",
    license="GPLv3",
    keywords="Algorithm Visualizer Animation Jupyter-notebook Graph",
    url="https://algviz.com/",
    project_urls={
        "Documentation": "https://algviz.readthedocs.io/en/latest/index.html",
        "Issue Tracker": "https://github.com/zjl9959/algviz/issues",
    },
    packages=['algviz'],
    install_requires=[
        'graphviz >= 0.8.4, != 0.18, <= 0.19.1',
        'ipykernel >= 6.4.0, <= 6.7.0'
    ],
    python_requires='>=3.7',
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Topic :: Scientific/Engineering :: Visualization",
        'Programming Language :: Python :: 3'
    ],
    zip_safe=False
)



#!/usr/bin/env python3

"""Define graph related classes and functions.

Author: zjl9959@gmail.com

License: GPLv3

"""

from algviz.utility import AlgvizRuntimeError, AlgvizParamError
from algviz.graph_node_base import GraphNodeBase as NodeBase


class GraphNeighborIter():
    """Neighbor nodes iterator for GraphNode class.

    """
    def __init__(self, node, neighbors):
        self._node = node
        self._neighbors = neighbors
        self._next_index = 0
    
    def __iter__(self):
        self._next_index = 0
        return self
    
    def __next__(self):
        if self._next_index >= len(self._neighbors):
            raise StopIteration
        else:
            neighbor = self._neighbors[self._next_index]
            self._next_index += 1
            return neighbor[0], neighbor[1]


class GraphNode(NodeBase):
    """This class defined the type of typology graph node.

    Attributes:
        val (printable): The label to be displayed in the graph node.
    """
    def __init__(self, val):
        super().__init__(val)
        super().__setattr__('_neighbors', list())   # tuple(Neighbor Node, Edge Label).


    def neighbors(self):
        """Return an iterator to iter over all the neighbor nodes of this node.
        
        Returns:
            GraphNeighborIter: Neighbor node iterator.
        """
        iter_neighbors = super().__getattribute__('_neighbors')
        return GraphNeighborIter(self, tuple(iter_neighbors))
    

    def neighborCount(self):
        """Return the neighbors count of this node.

        Returns:
            int: neighbors count.
        """
        neighbors_ = super().__getattribute__('_neighbors')
        return len(neighbors_)


    def neighborAt(self, index):
        """Return the neighbor node, edge at the index position of the neighbors list.
        
        Args:
            index (int): The position of the neighbor node.

        Returns:
            GraphNode: neighbor node.
            printable: the edge between this node and the neighbor node.
        
        Raises:
            AlgvizParamError: GraphNode neighbor index type error or out of range.
        """
        neighbors_ = super().__getattribute__('_neighbors')
        if type(index) != int or index < 0 or index >= len(neighbors_):
            raise AlgvizParamError('GraphNode neighbor index type error or out of range.')
        return neighbors_[index][0], neighbors_[index][1]


    def neighborIndex(self, node):
        """Return the neighbor node index in the neighbors list.

        Args:
            node (GraphNode): The neighbor node object to be located.

        Returns:
            int: the index position of the neighbor node. If node not found, then return -1.
        """
        res = -1
        neighbors_ = super().__getattribute__('_neighbors')
        for i in range(len(neighbors_)):
            if node == neighbors_[i][0]:
                res = i
        return res


    def add(self, node, edge=None, index=None):
        """Add a neighbor node for this node.
        
        Args:
            node (GraphNode): The neighbor node object to be added.
            edge (printable): The weight value of the edge between this node and neighbor node.
        
        Raises:
            AlgvizParamError: GraphNode neighbor index should be a positive integer!
        """
        if (type(index) != int and index != None) or (type(index) == int and index < 0):
            raise AlgvizParamError('GraphNode neighbor index should be a positive integer!')
        neighbors_ = super().__getattribute__('_neighbors')
        for (n, _) in neighbors_:
            if node == n:
                return
        if index == None or index >= len(neighbors_):
            neighbors_.append((node, edge))
        else:
            neighbors_.insert(index, (node, edge))
        self._on_update_neighbor_(node)


    def remove(self, node):
        """Remove one neighbor node. Do nothing if node not in neighbors collection.
        
        Args:
            node (GraphNode): The neighbor node to be removed.
        """
        neighbors_ = super().__getattribute__('_neighbors')
        for i in range(len(neighbors_)):
            if neighbors_[i][0] == node:
                neighbors_.pop(i)
                self._on_update_neighbor_(None)
                return


    def removeAt(self, index):
        """Remove one neighbor node. Do nothing if node not in neighbors collection.
        
        Args:
            index (int): The neighbor node index to be removed.

        Raises:
            AlgvizParamError: GraphNode neighbor index type error or out of range.
        """
        neighbors_ = super().__getattribute__('_neighbors')
        if type(index) != int or index < 0 or index >= len(neighbors_):
            raise AlgvizParamError('GraphNode neighbor index type error or out of range.')
        neighbors_.pop(index)
        self._on_update_neighbor_(None)


    def _neighbors_(self):
        """

        Returns:
            list[(neighbor_node, edge)]: All the neighbors nodes and edges.
        """
        neighbors_ = super().__getattribute__('_neighbors')
        return neighbors_


def updateGraphEdge(node1, node2, edge):
    """Update the graph's edge between node1 and node2.
    
    Args:
        node1,node2 (GraphNode): The node related to the updated edge.
        edge (printable): New weight value for this edge.
    """
    node1_neighbors = node1._neighbors
    pos_1 = node1.neighborIndex(node2)
    if pos_1 != -1:
        node1_neighbors[pos_1] = (node2, edge)
        node1_bind_graphs = node1.bind_graphs()
        for graph in node1_bind_graphs:
            graph._updateEdgeLabel(node1, node2, edge)
    node2_neighbors = node2._neighbors
    pos_2 = node2.neighborIndex(node1)
    if pos_2 != -1:
        node2_neighbors[pos_2] = (node1, edge)
        node2_bind_graphs = node2.bind_graphs()
        for graph in node2_bind_graphs:
            graph._updateEdgeLabel(node2, node1, edge)


def parseGraph(nodes, edges, nodes_label=None, directed=True):
    """Create a new graph from edges and nodes information.
    
    All the input edges's start and edge node should be found in nodes parameter.
    
    Example: parseGraph({0, 1, 2}, [(0, 1), (1, 2), (2, 0)], nodes_label={0:'node_0'})

    Args:
        nodes (set(printable)): All the nodes in this graph.
        edges (list(tuple(node1, node2, edge))): All the edges in this graph.
        nodes_label (dict(printable:printable)): Map the node id into it's display label.
        directed (bool): Should this graph be directed graph or undirected.

    Returns:
        dict(printable): All the graph nodes in this graph. Key is node label in nodes parameter, Value is GraphNode object.
    
    Raises:
        AlgvizRuntimeError: (paraseGraph) can not find node object for edge xxx.

    """
    # Create nodes for this graph.
    nodes_dict = dict()
    for node in nodes:
        if node in nodes_dict:
            continue
        node_val = node
        if nodes_label and node in nodes_label:
            node_val = nodes_label[node]
        nodes_dict[node] = GraphNode(node_val)
    # Add edges for this graph.
    for edge in edges:
        node1, node2, edge_val = edge[0], edge[1], None
        if node1 in nodes_dict and node2 in nodes_dict:
            if len(edge) == 3:
                edge_val = edge[2]
            nodes_dict[node1].add(nodes_dict[node2], edge_val)
            if directed == False:
                nodes_dict[node2].add(nodes_dict[node1], edge_val)
        else:
            raise AlgvizRuntimeError('(paraseGraph) can not find node object for edge {}.'.format(edge))
    return nodes_dict
